# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.
from http import HTTPStatus
from typing import Dict, Any

import requests
from requests import Response

from knowledge import logger
from knowledge.base.entity import LanguageCode


class ExtractionException(Exception):
    pass


def __extract_abstract__(title: str, language: str = 'en') -> str:
    """Extracting an abstract.

    Parameters
    ----------
    title: str -
        Title of wikipedia article
    language: str -
        language_code of Wikipedia

    Returns
    -------
    abstract: str
        Abstract of the wikipedia article
    """
    params: Dict[str, str] = {
        "action": "query",
        "format": "json",
        "titles": title,
        "prop": "extracts",
        "exintro": "1",
        "explaintext": "1",
        "redirects": "1"
    }

    url: str = f'https://{language}.wikipedia.org/w/api.php'
    response: Response = requests.get(url, params=params)
    if response.status_code == HTTPStatus.OK:
        result: Dict[str, Any] = response.json()
        if 'query' in result:
            pages = result['query']['pages']
            if len(pages) == 1:
                for v in pages.values():
                    return v.get('extract', '')
    raise ExtractionException(f"Abstract for article with {title} in language_code {language} cannot be extracted.")


def __extract_thumb__(title: str, language: str = 'en') -> str:
    """
    Extracting thumbnail from Wikipedia.

    Parameters
    ----------
    title: str
        Title of wikipedia article
    language: LanguageCode
        Language code of Wikipedia

    Returns
    -------
    url: str
        thumb URL
    """
    params: Dict[str, str] = {
        "action": "query",
        "format": "json",
        "titles": title,
        "prop": "pageimages",
        "pithumbsize": "400"
    }

    url: str = f'https://{language}.wikipedia.org/w/api.php'
    response: Response = requests.get(url, params=params)
    if response.ok:
        result: dict = response.json()
        try:
            if 'query' in result:
                pages: dict = result['query']['pages']
                if len(pages) == 1:
                    for v in pages.values():
                        if 'thumbnail' in v:
                            return v['thumbnail']['source']
        except Exception as e:
            logger.error(e)
    raise ExtractionException(f"Thumbnail for article with {title} in language_code {language} cannot be extracted.")


def get_wikipedia_summary(title: str, lang: str = 'en') -> Dict[str, str]:
    """
    Extracting summary image and abstract for wikipedia URL.

    Parameters
    ----------
    title: str
        Title of the Wikipedia article
    lang: str
        Language code

    Returns
    -------
    result: Dict[str, str]
        Summary dict with image and summary text
    """
    try:
        thumbnail: str = __extract_thumb__(title, lang)
    except ExtractionException as _:
        thumbnail = ''
    try:
        summary: str = __extract_abstract__(title, lang)
    except ExtractionException as _:
        summary = ''
    return {
        'summary-image': thumbnail,
        'summary-text': summary
    }


def get_wikipedia_summary_url(wiki_url: str, lang: str = 'en') -> Dict[str, str]:
    """
    Extracting summary image and abstract for wikipedia URL.
    Parameters
    ----------
    wiki_url: str
        Wikipedia URL
    lang: str
        Language code

    Returns
    -------
    result: Dict[str, str]
        Result dictionary.
    """
    title: str = wiki_url.split('/')[-1]
    return {
        'url': wiki_url,
        'summary-image': __extract_thumb__(title, lang),
        'summary-text': __extract_abstract__(title, lang)
    }



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.

__all__ = ['base', 'graph', 'ontology']

from knowledge.services import base
from knowledge.services import graph
from knowledge.services import ontology



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.
import hashlib
import math
import urllib.parse
from abc import ABC
from enum import Enum
from http import HTTPStatus
from typing import Tuple, List, Dict, Set, Any

import dateutil.parser
import requests
from qwikidata.entity import WikidataItem, WikidataClaimGroup, WikidataProperty
from qwikidata.linked_data_interface import get_entity_dict_from_api, InvalidEntityId
from qwikidata.sparql import get_subclasses_of_item
from qwikidata.typedefs import PropertyId, ItemId, LanguageCode

from knowledge.public import PROPERTY_MAPPING
from knowledge.utils.wikipedia import get_wikipedia_summary, get_wikipedia_summary_url


class WikiDataAPIException(Exception):
    pass


# OntologyPropertyReference constants
INSTANCE_OF_PROPERTY: str = 'P31'
IMAGE_PROPERTY: str = 'P18'
API_LIMIT: int = 50

THUMB_IMAGE_URL: str = 'https://upload.wikimedia.org/wikipedia/commons/thumb/{}/{}/{}/200px-{}'
MULTIPLE_ENTITIES_API: str = 'https://www.wikidata.org/w/api.php?action=wbgetentities&ids={}&languages={}&format=json'


class Precision(Enum):
    """
    Precision enum for date.
    """
    BILLION_YEARS = 0
    MILLION_YEARS = 3
    HUNDREDS_THOUSAND_YEARS = 4
    MILLENIUM = 6
    CENTURY= 7
    DECADE = 8
    YEAR = 9
    MONTH = 10
    DAY = 11


# Wikidata Properties
STUDENT_OF: str = 'P1066'
STUDENT: str = 'P802'
INCEPTION: str = 'P571'
MOVEMENT: str = 'P135'
SUBCLASS_OF: str = 'P279'
TITLE: str = 'P1476'
COLLECTION: str = 'P195'
GENRE: str = 'P136'
CREATOR: str = 'P170'
LOGO_IMAGE: str = 'P154'
FLAG_IMAGE: str = 'P41'
GREGORIAN_CALENDAR: str = 'Q1985727'
START_TIME: str = 'P580'
END_TIME: str = 'P582'
FOLLOWS: str = 'P155'
FOLLOWED_BY: str = 'P156'
COUNTRY_OF_ORIGIN: str = 'P495'
COUNTRY: str = 'P17'
INSTANCE_OF: str = 'P31'
IMAGE: str = 'P18'
# URL - Wikidata
GREGORIAN_CALENDAR_URL: str = 'http://www.wikidata.org/entity/Q1985786'
# Template of wikidata entity
WIKIDATA_ORG_WIKI_TEMPLATE: str = 'https://www.wikidata.org/wiki/{}'
# URL - Wikidata service
WIKIDATA_SPARQL_URL: str = "https://query.wikidata.org/sparql"


def wikidate(param: Dict[str, Any]) -> dict:
    """
    Parse and extract wikidata structure.
    Parameters
    ----------
    param: Dict[str, Any]
        Entity wikidata

    Returns
    -------
    result: Dict[str, Any]
        Dict with pretty print of date
    """
    time: str = param['values'][0]['time']
    timezone: int = param['values'][0]['timezone']
    before: int = param['values'][0]['before']
    after: int = param['values'][0]['after']
    precision: int = param['values'][0]['precision']
    calendar_model: str = param['values'][0]['calendarmodel']
    after_christ: bool = True
    pretty: str = ''
    if time.startswith('+'):
        time = time[1:]
    elif time.startswith('-'):
        time = time[1:]
        after_christ = False
    # Probably not necessary
    date_str = time.strip()
    # Remove + sign
    if date_str[0] == '+':
        date_str = date_str[1:]
    # Remove missing month/day
    date_str = date_str.split('-00', maxsplit=1)[0]
    # Parse date
    dt = dateutil.parser.parse(date_str)
    if Precision.BILLION_YEARS.value == precision:
        pass
    elif Precision.MILLION_YEARS.value == precision:
        pass
    elif Precision.HUNDREDS_THOUSAND_YEARS.value == precision:
        pass
    elif Precision.MILLENIUM.value == precision:
        pass
    elif Precision.CENTURY.value == precision:
        century: int = int(math.ceil(dt.year / 100))
        pretty = '{}th century'.format(century)
    elif Precision.DECADE.value == precision:
        pretty = '{}s{}'.format(dt.year, '' if after_christ else ' BC')
    elif Precision.YEAR.value == precision:
        pretty = '{}{}'.format(dt.year, '' if after_christ else ' BC')
    elif Precision.MONTH.value == precision:
        pretty = dt.strftime("%B %Y")
    elif Precision.DAY.value == precision:
        pretty = dt.strftime("%-d %B %Y")
    return {
        'time': time,
        'timezone': timezone,
        'before': before,
        'after': after,
        'precision': precision,
        'calendarmodel': calendar_model,
        'pretty': pretty
    }


class WikiDataAPIClient(ABC):
    """WikiData API client.
    """
    ACTIVATION_RELATIONS = [
        'P625',  # coordinate location',
        'P31',  # instance of',
        'P527',  # 'has part',
        'P361',  # 'part of',
        'P856',  # 'official website',
        'P47',  # 'shares border with',
        'P6',  # 'head of government',
        'P1082',  # 'population',
        'P421',  # 'located in timezone',
        'P17',  # 'country',
        'P910',  # 'topics main category',
        'P373',  # 'Commons category',
        'P901',  # 'FIPS 10-4 (countries and regions)',
        'P948',  # 'page banner',
        'P646',  # 'Freebase ID',
        'P1376',  # 'capital of',
        'P1814',  # 'name in kana',
        'P571',  # 'inception',
        'P2046',  # 'area',
        'P163',  # 'flag',
        'P2044',  # 'elevation above sea level',
        'P610',  # 'highest point',
        'P1365',  # 'replaces',
        'P36',  # 'capital',
        'P6794',  # 'minimum wage',
        'P18',  # 'image',
        # Person
        'P21',  # 'sex or gender',
        'P345',  # 'IMDb ID',
        'P27',  # 'country of citizenship',
        'P19',  # 'place of birth',
        'P569',  # 'date of birth',
        'P106',  # 'occupation',
        'P102',  # 'political party',
        'P166',  # 'award received',
        'P2002',  # 'Twitter username',
        'P26',  # 'spouse',
        'P735',  # 'given name',
        'P734',  # 'family name',
        'P1412',  # 'languages spoken, written or signed',
        'P69',  # 'educated at',
        'P1559',  # 'name in native language_code',
        'P1477',  # 'birth name',
        'P103',  # 'native language_code',
        'P793',  # 'significant event',
        'P2003',  # 'Instagram username',
        'P1340',  # 'eye color',
        'P1884',  # 'hair color',
        'P2048',  # 'height',
        'P2067',  # 'mass',
        'P108',  # 'employer',
        'P800',  # 'notable work',
        'P10',  # 'video',
        'P1971',  # 'number of children',
        'P91'  # 'sexual orientation'
    ]

    PROPERTY_OFFICIAL_WEBSITE: str = 'P856'
    PROPERTY_PERSON_FIRSTNAME: str = 'P735'
    PROPERTY_PERSON_LASTNAME: str = 'P734'

    def __init__(self):
        self.PROPERTIES = list(WikiDataAPIClient.ACTIVATION_RELATIONS)

    @staticmethod
    def image_url(img: str):
        extension: str = ''
        conversion: str = ''
        fixed_img: str = img.replace(' ', '_')
        if fixed_img.lower().endswith('svg'):
            extension: str = '.png'
        if fixed_img.lower().endswith('tif') or fixed_img.lower().endswith('tiff'):
            extension: str = '.jpg'
            conversion: str = 'lossy-page1-'
        hash_img = hashlib.md5(fixed_img.encode('utf-8')).hexdigest()
        url_img_part: str = urllib.parse.quote_plus(fixed_img)
        return THUMB_IMAGE_URL.format(hash_img[0], hash_img[:2], url_img_part, conversion + url_img_part + extension)

    @staticmethod
    def wikipedia_url(wikidata_id: str, lang: str = 'en'):
        url: str = 'https://www.wikidata.org/w/api.php?action=wbgetentities&props=sitelinks/urls&ids={}&format=json' \
            .format(wikidata_id)
        json_response = requests.get(url).json()
        entities: dict = json_response['entities']
        if entities:
            entity: dict = entities[wikidata_id]
            if entity:
                sitelinks: dict = entity['sitelinks']
                if sitelinks:
                    # filter only the specified language_code
                    sitelink: str = sitelinks.get(f'{lang}wiki')
                    if sitelink:
                        wiki_url = sitelink.get('url')
                        if wiki_url:
                            return requests.utils.unquote(wiki_url)
        return None

    @staticmethod
    def __entity_image__(entity: WikidataItem):
        p: WikidataClaimGroup = entity.get_claim_group(PropertyId(IMAGE_PROPERTY))
        if len(p) > 0 and p[0].mainsnak.datavalue:
            img: str = p[0].mainsnak.datavalue.value.replace(' ', '_')
            return WikiDataAPIClient.image_url(img)
        return ''

    @staticmethod
    def __entity_image_dict__(prop: List[dict]):
        if len(prop) > 0:
            img: str = prop[0]['mainsnak']['datavalue']['value'].replace(' ', '_')
            return WikiDataAPIClient.image_url(img)
        return ''

    @staticmethod
    def __entity_type__(entity: WikidataItem) -> str:
        p: WikidataClaimGroup = entity.get_claim_group(PropertyId('P31'))
        if len(p) > 0:
            type_entity = p[0].mainsnak.datavalue.value['id']
            return type_entity
        return ''

    @staticmethod
    def __entity_type_dict__(entity: dict) -> set:
        if INSTANCE_OF_PROPERTY in entity:
            props: dict = entity[INSTANCE_OF_PROPERTY]
            if len(props) > 0:
                return set([p['mainsnak']['datavalue']['value']['id'] for p in props])
        return set()

    @staticmethod
    def __entity_type_obj__(entity: WikidataItem) -> Set[str]:
        props: WikidataClaimGroup = entity.get_claim_group(PropertyId(INSTANCE_OF_PROPERTY))
        if len(props) > 0:
            return set([p.mainsnak.datavalue.value['id'] for p in props])
        return set()

    @staticmethod
    def __data_type__(prop: list):
        return prop[0]['mainsnak']['datatype']

    @staticmethod
    def __entity_id__(prop: list):
        if len(prop) > 0:
            if 'datavalue' in prop[0]['mainsnak']:
                return prop[0]['mainsnak']['datavalue']['value']['id']
        return None

    @staticmethod
    def __value__(prop: list):
        data_type: str = WikiDataAPIClient.__data_type__(prop)
        val = None
        if 'datavalue' in prop[0]['mainsnak']:
            if data_type == 'monolingualtext':
                val = prop[0]['mainsnak']['datavalue']['value']['text']
            elif data_type in ['string', 'external-id', 'url']:
                val = prop[0]['mainsnak']['datavalue']['value']
            elif data_type == 'commonsMedia':
                val = WikiDataAPIClient.__entity_image_dict__(prop)
            elif data_type == 'time':
                val = prop[0]['mainsnak']['datavalue']['value']['time']
            elif data_type == 'quantity':
                if 'amount' in prop[0]['mainsnak']['datavalue']['value']:
                    val = '{} {}'.format(prop[0]['mainsnak']['datavalue']['value']['amount'],
                                         prop[0]['mainsnak']['datavalue']['value']['unit'])

            elif data_type == 'wikibase-lexeme':
                val = prop[0]['mainsnak']['datavalue']['value']['id']
            elif data_type in ['geo-shape', 'wikibase-property']:
                # Not supported
                val = None
            elif data_type == 'globe-coordinate':
                val = {
                    'longitude': prop[0]['mainsnak']['datavalue']['value']['longitude'],
                    'latitude': prop[0]['mainsnak']['datavalue']['value']['latitude'],
                    'altitude': prop[0]['mainsnak']['datavalue']['value']['altitude']
                }
            else:
                raise WikiDataAPIException(f"Unkown type {data_type}")
        return val

    @staticmethod
    def __entity_to_dict__(entity: WikidataItem, language: LanguageCode = LanguageCode('en')) -> Dict[str, Any]:
        type_uris: set = WikiDataAPIClient.__entity_type_obj__(entity)
        return {
            'uri': entity.entity_id,
            'label': entity.get_label(language),
            'description': entity.get_description(language),
            'alias': entity.get_aliases(language),
            'image': WikiDataAPIClient.__entity_image__(entity),
            'types': type_uris
        }

    @staticmethod
    def __entity_to_dict_lang__(entity: WikidataItem, languages: List[LanguageCode]) -> Dict[str, Any]:
        type_uris: Set[str] = WikiDataAPIClient.__entity_type_obj__(entity)
        image: str = WikiDataAPIClient.__entity_image__(entity)
        return {
            'uri': entity.entity_id,
            'label': dict([(lang, entity.get_label(lang)) for lang in languages]),
            'description': dict([(lang, entity.get_description(lang)) for lang in languages]),
            'alias': dict([(lang, entity.get_aliases(lang)) for lang in languages]),
            'image': image,
            'types': type_uris
        }

    @staticmethod
    def __pull_entities__(language: str, uris: list):
        query = '|'.join(map(str, uris))
        response = requests.get(MULTIPLE_ENTITIES_API.format(query, language))
        if response.status_code != HTTPStatus.OK:
            raise WikiDataAPIException('Response return has not been successful. [HTTP Code:={}]'
                                       .format(response.status_code))
        return response.json()

    @staticmethod
    def __label__(v: dict, language: str):
        if 'labels' in v and language in v['labels']:
            return v['labels'][language]['value']
        return ''

    @staticmethod
    def __description__(v: dict, language: str):
        if 'descriptions' in v and language in v['descriptions']:
            return v['descriptions'][language]['value']
        return ''

    @staticmethod
    def __extract_entities__(result: dict, language: str, add_literals: bool = True, add_relations: bool = False):
        relations: list = []
        entities: list = []
        missing_entities: set = set()
        for qid, v in result['entities'].items():
            # Extracting the properties
            if IMAGE_PROPERTY in v['claims']:
                image_url: str = WikiDataAPIClient.__entity_image_dict__(v['claims'][IMAGE_PROPERTY])
            else:
                image_url: str = ''
            type_uris: set = WikiDataAPIClient.__entity_type_dict__(v['claims'])
            label: str = WikiDataAPIClient.__label__(v, language)
            description: str = WikiDataAPIClient.__description__(v, language)
            # Collect the literals for entity
            literals: dict = {}
            # Now iterate over claims
            for pid, value in v['claims'].items():
                if pid in WikiDataAPIClient.ACTIVATION_RELATIONS:
                    if WikiDataAPIClient.__data_type__(value) == 'wikibase-item':
                        # Extract the relations
                        if add_relations:
                            t_qid = WikiDataAPIClient.__entity_id__(value)
                            if pid in PROPERTY_MAPPING and t_qid:
                                missing_entities.add(t_qid)
                                relations.append({'property': pid,
                                                  'subject': qid, 'predicate': PROPERTY_MAPPING[pid], 'object': t_qid})
                    else:
                        # Extract the literals
                        if add_literals:
                            if pid in PROPERTY_MAPPING:
                                # Collect literals
                                literals[pid] = {
                                    'uri': pid,
                                    'label': PROPERTY_MAPPING[pid],
                                    'value': WikiDataAPIClient.__value__(value)
                                }

            # Adding entity to result list
            entities.append({'uri': qid, 'label': label, 'description': description,
                             'image': image_url,
                             'literals': literals})
        return entities, relations, missing_entities

    def activations(self, uris: list, language: str = 'en') -> tuple:
        """Activations of URIs
        :param uris: list of URIs
        :param language: language_code of entity
        :return: list of entities, list of relations
        """
        result = self.__pull_entities__(language, uris)
        entities, relations, missing_entities = WikiDataAPIClient.__extract_entities__(result, language,
                                                                                       add_literals=True,
                                                                                       add_relations=True)
        jobs = list(missing_entities)
        # pull missing entities
        while len(jobs) > 0:
            result_other = self.__pull_entities__(language, jobs[:API_LIMIT])
            more_entities, _, _ = WikiDataAPIClient.__extract_entities__(result_other, language,
                                                                         add_literals=True,
                                                                         add_relations=False)
            entities.extend(more_entities)
            del jobs[:API_LIMIT]
        return entities, relations

    def entity(self, qid: str, language: str = 'en', pull_wiki_content: bool = False) -> dict:
        """Get entity information from public including relations

        :param qid: QID representing the entity in the public knowledge graph
        :param language: language_code for text
        :param pull_wiki_content: pulling extended description and summary
        :return: dict with relevant information
        """
        entity_dict = get_entity_dict_from_api(ItemId(qid))
        entity = WikidataItem(entity_dict)
        entity_dict: dict = self.__entity_to_dict__(entity, language=LanguageCode(language))
        if pull_wiki_content:
            url: str = WikiDataAPIClient.wikipedia_url(qid)
            if url:
                summary: dict = get_wikipedia_summary_url(url)
                entity_dict['image'] = summary['summary-image']
                entity_dict['description'] = summary['summary-text']
                entity_dict['wikiurl'] = url
        return entity_dict

    def entity_lang(self, qid: ItemId, languages: List[LanguageCode] = None, pull_wiki_content: bool = False) -> dict:
        """Get entity information from public including relations

        :param qid: str -
         QID representing the entity in the public knowledge graph
        :param languages: List[str] -
            List of languages for text
        :param pull_wiki_content: pulling extended description and summary
        :return: dict with relevant information
        """
        entity_dict = get_entity_dict_from_api(qid)
        entity = WikidataItem(entity_dict)
        entity_dict: dict = self.__entity_to_dict_lang__(entity, languages=languages)
        if pull_wiki_content:
            entity_dict['wikiurl'] = {}
            for lang in languages:
                url: str = WikiDataAPIClient.wikipedia_url(qid, lang)
                if url:
                    summary: dict = get_wikipedia_summary_url(url, lang)
                    if len(summary['summary-image']) > 0:
                        entity_dict['image'] = summary['summary-image']
                    if len(summary['summary-text']) > 0:
                        entity_dict['description'][lang] = summary['summary-text']
                    entity_dict['wikiurl'][lang] = url
        return entity_dict

    def entity_rels(self, qid: str, language: str = 'en', pull_wiki_content: bool = False) -> Tuple[dict, dict]:
        """Get entity information from public including relations

        :param qid: QID representing the entity in the public knowledge graph
        :param language: language_code for text
        :param pull_wiki_content: pulling extended description and summary
        :return: dict with relevant information
        """
        entity_dict: dict = get_entity_dict_from_api(ItemId(qid))
        entity: WikidataItem = WikidataItem(entity_dict)
        properties: dict = {}
        for p, claim_group in entity.get_truthy_claim_groups().items():
            literal = [
                claim.mainsnak.datavalue.value
                for claim in claim_group
                if claim.mainsnak.snaktype == "value"
            ]
            properties[p] = {
                'property-label': WikiDataAPIClient.property(p, LanguageCode(language)),
                'values': literal
            }
        entity_result: dict = self.__entity_to_dict__(entity, language=LanguageCode(language))
        wikiurl: dict = entity.get_sitelinks(f'{language}wiki')
        if len(wikiurl) > 0 and f'{language}wiki' in wikiurl:
            url: str = wikiurl[f'{language}wiki']['url']
            title: str = wikiurl[f'{language}wiki']['title']
            entity_result['wikiurl'] = url
            if pull_wiki_content:
                summary: dict = get_wikipedia_summary(title)
                # Only update if there is a meaningful update
                img: str = summary['summary-image']
                if img:
                    entity_result['image'] = img
                abstract: str = summary['summary-text']
                if abstract:
                    entity_result['description'] = abstract
        return entity_result, properties

    def entity_rels_lang(self, qid: str, languages: List[LanguageCode] = None, pull_wiki_content: bool = False,
                         default_language: str = 'en') -> tuple:
        """Get entity information from public including relations.

        :param qid: str -
            QID representing the entity in the public knowledge graph
        :param languages: List[str] -
            List of languages for text
        :param pull_wiki_content: bool -
            Pulling extended description and summary
        :param default_language: str -
            Default language_code
        :return: dict with relevant information
        """
        entity_dict: dict = get_entity_dict_from_api(ItemId(qid))
        entity: WikidataItem = WikidataItem(entity_dict)
        properties: Dict[str, dict] = {}
        # Properties
        for p, claim_group in entity.get_truthy_claim_groups().items():
            literal: List = [
                claim.mainsnak.datavalue.value
                for claim in claim_group
                if claim.mainsnak.snaktype == "value"
            ]
            properties[p] = {
                'property-label': WikiDataAPIClient.property(p, default_language),
                'values': literal
            }
        entity_result: dict = self.__entity_to_dict_lang__(entity, languages=languages)
        for lang in languages:
            wiki_url: dict = entity.get_sitelinks(f'{lang}wiki')
            if len(wiki_url) > 0 and f'{lang}wiki' in wiki_url:
                if 'wikiurl' not in entity_result:
                    entity_result['wikiurl'] = {}
                url: str = wiki_url[f'{lang}wiki']['url']
                title: str = wiki_url[f'{lang}wiki']['title']
                entity_result['wikiurl'][lang] = url
                if pull_wiki_content:
                    summary: dict = get_wikipedia_summary(title, lang)
                    # Only update if there is a meaningful update
                    img: str = summary['summary-image']
                    if img is not None or img == '':
                        entity_result['image'] = img
                    abstract: str = summary['summary-text']
                    if abstract is not None or abstract == '':
                        entity_result['description'][lang] = abstract
        return entity_result, properties

    def references(self, qid: str, language: str = 'en'):
        references = []
        entity, properties = self.entity_rels(qid, language)
        if WikiDataAPIClient.PROPERTY_OFFICIAL_WEBSITE in properties:
            references.append({'property': properties[WikiDataAPIClient.PROPERTY_OFFICIAL_WEBSITE],
                               'property-label': PROPERTY_MAPPING[WikiDataAPIClient.PROPERTY_OFFICIAL_WEBSITE]})
        return references

    def entities_rel(self, qids: list, language: str = 'en') -> tuple:
        """Return relations and entities for qids.

        :param qids: list of entities in Wikidata
        :param language: language_code the content
        :return: entities, relations
        """
        jobs = list(qids)
        entities: list = []
        relations: list = []
        while len(jobs) > 0:
            result = self.__pull_entities__(language, jobs[:API_LIMIT])
            p_entities, p_relations, _ = WikiDataAPIClient.__extract_entities__(result, language,
                                                                                add_literals=True,
                                                                                add_relations=True)
            del jobs[:API_LIMIT]
            entities.extend(p_entities)
            relations.extend(p_relations)

        return entities, relations

    @staticmethod
    def property(pid: str, language: str = 'en'):
        """Extract property.

        :param pid: property id
        :param language: language_code code for property name
        :return: property nam
        """
        try:
            if pid in PROPERTY_MAPPING:  # only English mappings
                return PROPERTY_MAPPING[pid]
            else:
                prop_dict = get_entity_dict_from_api(PropertyId(pid))
                return WikidataProperty(prop_dict).get_label(LanguageCode(language))
        except InvalidEntityId as e:
            return None

    @staticmethod
    def subclasses(qid: str) -> list:
        """Check for subclasses.

        :param qid: QID of entity
        :return: list of subclasses QIDs
        """
        subclasses_of_uri = get_subclasses_of_item(qid)
        return subclasses_of_uri



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.
from json import JSONEncoder
from typing import List, Union, Set, Tuple

from knowledge import logger
from knowledge.base.access import TenantAccessRight
from knowledge.base.entity import *


# ---------------------------------------------- Vocabulary base URI ---------------------------------------------------
PREFIX: str = "xsd"
BASE_URI: str = "http://www.w3.org/2001/XMLSchema#"
# ---------------------------------------------------- Constants -------------------------------------------------------
SUB_CLASS_OF_TAG: str = 'subClassOf'
TENANT_ID: str = 'tenantId'
NAME_TAG: str = "name"


class PropertyType(enum.Enum):
    """
    PropertyType
    -----------
    Within the ontology two different property types are defined. A data- and an object property.
    """
    OBJECT_PROPERTY = "Relation"
    DATA_PROPERTY = "Literal"


INVERSE_PROPERTY_TYPE: Dict[str, PropertyType] = dict([(pt.value, pt) for pt in PropertyType])


class DataPropertyType(enum.Enum):
    """
    DataPropertyType.
    -----------------
    Data types that are used by Datatype properties.
    """
    STRING = BASE_URI + "string"
    """Character strings (but not all Unicode character strings) """
    BOOLEAN = BASE_URI + "boolean"
    """boolean: true, false"""
    DECIMAL = BASE_URI + "decimal"
    """Arbitrary-precision decimal numbers"""
    INTEGER = BASE_URI + "integer"
    """Arbitrary-size integer numbers"""
    DOUBLE = BASE_URI + "double"
    """64-bit floating point numbers incl. ±Inf, ±0, NaN"""
    FLOAT = BASE_URI + "float"
    """32-bit floating point numbers incl. ±Inf, ±0, NaN"""
    DATE = BASE_URI + "date"
    """Dates (yyyy-mm-dd) with or without timezone"""
    TIME = BASE_URI + "time"
    """Times (hh:mm:ss.sss…) with or without timezone"""
    DATE_TIME = BASE_URI + "dateTime"
    """Date and time with or without timezone"""
    DATE_TIMESTAMP = BASE_URI + "dateTimeStamp"
    """Date and time with required timezone """
    G_YEAR = BASE_URI + "gYear"
    """Gregorian calendar year"""
    G_MONTH = BASE_URI + "gMonth"
    """Gregorian calendar month"""
    G_DAY = BASE_URI + "gDay"
    """Gregorian calendar day of the month"""
    G_YEAR_MONTH = BASE_URI + "gYearMonth"
    """Gregorian calendar year and month"""
    G_MONTH_DAY = BASE_URI + "gMonthDay"
    """Gregorian calendar month and day"""
    DURATION = BASE_URI + "duration"
    """Duration of time"""
    YEAR_MONTH_DURATION = BASE_URI + "yearMonthDuration"
    """Duration of time (months and years only)"""
    DAYTIME_DURATION = BASE_URI + "dayTimeDuration"
    """Duration of time (days, hours, minutes, seconds only)"""
    BYTES = BASE_URI + "byte"
    """-128…+127 (8 bit)"""
    SHORT = BASE_URI + "short"
    """-32768… + 32767 (16 bit)"""
    INT = BASE_URI + "int"
    """-2147483648…+2147483647 (32 bit)"""
    LONG = BASE_URI + "long"
    """-9223372036854775808…+9223372036854775807 (64 bit)"""
    UNSIGNED_BYTE = BASE_URI + "unsignedByte"
    """0 … 255 (8 bit)"""
    UNSIGNED_SHORT = BASE_URI + "unsignedShort"
    """0 … 65535 (16 bit)"""
    UNSIGNED_INT = BASE_URI + "unsignedInt"
    """ 0 … 4294967295 (32 bit)"""
    UNSIGNED_LONG = BASE_URI + "unsignedLong"
    """  0 … 18446744073709551615 (64 bit)"""
    POSITIVE_INTEGER = BASE_URI + "positiveInteger"
    """Integer numbers > 0 """
    NON_NEGATIVE_INTEGER = BASE_URI + "nonNegativeInteger"
    """Integer numbers ≥ 0"""
    NEGATIVE_INTEGER = BASE_URI + "negativeInteger"
    """Integer numbers ≤ 0"""
    NON_POSITIVE_INTEGER = BASE_URI + "nonPositiveInteger"
    """Integer numbers ≤ 0"""
    HEX_BINARY = BASE_URI + "hexBinary"
    """Hex-encoded binary data"""
    BASE64_BINARY = BASE_URI + "base64Binary"
    """Base64-encoded binary data"""
    ANY_URI = BASE_URI + "anyURI"
    """Absolute or relative URIs and IRIs"""
    LANGUAGE = BASE_URI + "language_code"
    """Language tags per http://tools.ietf.org/html/bcp47"""
    NORMALIZED = BASE_URI + "normalizedString"
    """Whitespace-normalized strings"""
    TOKEN = BASE_URI + "token"
    """Tokenized strings"""
    NM_TOKEN = BASE_URI + "NMTOKEN"
    """XML NMTOKENs"""
    NAME = BASE_URI + "Name"
    """XML Names"""
    NC_NAME = BASE_URI + "NCName"
    """XML NCNames"""


INVERSE_DATA_PROPERTY_TYPE_MAPPING: Dict[str, DataPropertyType] = dict([(lit_type.value, lit_type)
                                                                        for lit_type in DataPropertyType])
"""Maps the string representation of the XSD data types to the data types enum constants."""


# ------------------------------------------ Ontology References -------------------------------------------------------

class OntologyObjectReference(abc.ABC):
    """
        Ontology class type
        ------------------
        Associated to an entity to link the type of the entity.

        Parameters
        ----------
        scheme: str
            Scheme or owner of the ontology object
        context: str
            Context of ontology object
        name: str
            Ontology object reference name
    """

    def __init__(self, scheme: str, context: str, name: str):
        self.__scheme: str = scheme
        self.__context: str = context
        self.__name: str = name

    @property
    def scheme(self):
        """Scheme."""
        return self.__scheme

    @property
    def context(self):
        """Context."""
        return self.__context

    @property
    def name(self):
        """Name."""
        return self.__name

    @property
    def iri(self):
        """Internationalized Resource Identifier (IRI) encoded ontology class name."""
        return f'{self.scheme}:{self.context}#{self.name}'

    def __repr__(self):
        return self.iri

    @classmethod
    def parse_iri(cls, iri: str) -> Tuple[str, str, str]:
        colon_idx: int = iri.index(':')
        hash_idx: int = iri.index('#')
        scheme: str = iri[:colon_idx]
        context: str = iri[colon_idx + 1:hash_idx]
        name: str = iri[hash_idx + 1:]
        return scheme, context, name


class OntologyClassReference(OntologyObjectReference):
    """
    Ontology class type
    -------------------
    Associated to an ontology class.

    Parameters
    ----------
    scheme: str
        Scheme or owner
    context: str
        Context of class
    class_name: str
        Class name
    """

    def __init__(self, scheme: str, context: str, class_name: str):
        super().__init__(scheme, context, class_name)

    @property
    def class_name(self):
        """Class name."""
        return self.name

    @classmethod
    def parse(cls, iri: str) -> 'OntologyClassReference':
        scheme, context, name = OntologyObjectReference.parse_iri(iri)
        return OntologyClassReference(scheme, context, name)

    def __eq__(self, other):
        if not isinstance(other, OntologyClassReference):
            return False
        return self.iri == other.iri

    def __hash__(self):
        return hash(self.iri)


class OntologyPropertyReference(OntologyObjectReference):
    """
    Property reference
    ------------------
    Associated to an ontology property.

    Parameters
    ----------
    scheme: str
        Scheme or owner
    context: str
        Context of class
    property_name: str
        Property name
    """

    def __init__(self, scheme: str, context: str, property_name: str):
        super().__init__(scheme, context, property_name)

    @property
    def property_name(self):
        """Property name."""
        return self.name

    @classmethod
    def parse(cls, iri: str) -> 'OntologyPropertyReference':
        scheme, context, name = OntologyObjectReference.parse_iri(iri)
        return OntologyPropertyReference(scheme, context, name)

    def __eq__(self, other):
        if not isinstance(other, OntologyPropertyReference):
            return False
        return self.iri == other.iri

    def __hash__(self):
        return hash(self.iri)


# ------------------------------------------------- Constants ----------------------------------------------------------
THING_CLASS: OntologyClassReference = OntologyClassReference('wacom', 'core', 'Thing')
SYSTEM_SOURCE_SYSTEM: OntologyPropertyReference = OntologyPropertyReference('wacom', 'core', 'sourceSystem')
SYSTEM_SOURCE_REFERENCE_ID: OntologyPropertyReference = OntologyPropertyReference('wacom', 'core', 'sourceReferenceId')


class OntologyClass(OntologyObject):
    """
    OntologyClass
    ----------------
    Concept for ontology.

    Parameters
    ----------
    tenant_id: str
        Tenant id for ontology
    context: str
        Context
    reference: OntologyClassReference
        Reference for ontology class
    icon: str
        Icon representing concept
    labels: List[Label]
        List of labels
    comments: List[Comment]
        List of comments

    subclass_of: str (default: None)
        Subclass of ontology class
    """

    def __init__(self, tenant_id: str, context: str, reference: OntologyClassReference,
                 subclass_of: OntologyClassReference = None, icon: Optional[str] = None,
                 labels: Optional[List[Label]] = None, comments: Optional[List[Comment]] = None):
        self.__subclass_of: OntologyClassReference = subclass_of
        self.__reference: OntologyClassReference = reference
        super().__init__(tenant_id, reference.iri, icon, labels, comments, context)

    @property
    def subclass_of(self) -> Optional[OntologyClassReference]:
        """
        Superclass of the class.
        """
        return self.__subclass_of

    @property
    def reference(self) -> OntologyClassReference:
        """
        Reference of ontology class.
        """
        return self.__reference

    def __repr__(self):
        return f'<OntologyClass> - [reference:={self.reference}, subclass_of:={self.subclass_of}]'

    @classmethod
    def from_dict(cls, concept_dict: Dict[str, Any]):
        labels: List[Label] = [] if concept_dict[LABELS_TAG] is None else \
            [Label(content=la[VALUE_TAG], language_code=la[LANGUAGE_CODE_TAG]) for la in concept_dict[LABELS_TAG]]
        comments: List[Comment] = [] if concept_dict['comments'] is None else \
            [Comment(text=la[VALUE_TAG], language_code=la[LANGUAGE_CODE_TAG]) for la in concept_dict['comments']]
        return OntologyClass(tenant_id=concept_dict[TENANT_ID], context=concept_dict['context'],
                             reference=OntologyClassReference.parse(concept_dict[NAME_TAG]),
                             subclass_of=OntologyClassReference.parse(concept_dict[SUB_CLASS_OF_TAG]),
                             icon=concept_dict['icon'], labels=labels, comments=comments)

    @classmethod
    def new(cls) -> 'OntologyClass':
        return OntologyClass('', '', THING_CLASS)


class OntologyProperty(OntologyObject):
    """
    Ontology Property
    -----------------
    Property ontology object.

    Parameters
    ----------
    kind: str
        Kind of relation
    tenant_id: str
        Tenant id
    context: str
        Context
    name: OntologyPropertyReference
        Name of property object
    icon: str
        Icon describing the property
    property_domain: OntologyClassReference
        Domain for the property
    property_range: OntologyClassReference
        Range for the property
    labels: List[Label]
        List of labels (localized)
    comments: List[Comment],
        List of comments
    subproperty_of: str (default: = None)
        Subproperty
    inverse_property_of: str (optional)
        Inverse property
    """

    def __init__(self, kind: PropertyType, tenant_id: str, context: str, name: OntologyPropertyReference,
                 icon: str = None,
                 property_domain: Optional[OntologyClassReference] = None,
                 property_range: Optional[Union[OntologyClassReference, DataPropertyType]] = None,
                 labels: Optional[List[Label]] = None,
                 comments: Optional[List[Comment]] = None,
                 subproperty_of: Optional[OntologyPropertyReference] = None,
                 inverse_property_of: Optional[OntologyPropertyReference] = None):
        self.__kind: PropertyType = kind
        self.__subproperty_of: OntologyPropertyReference = subproperty_of
        self.__inverse_property_of: OntologyPropertyReference = inverse_property_of
        self.__domain: OntologyClassReference = property_domain
        self.__range: Optional[Union[OntologyClassReference, DataPropertyType]] = property_range
        self.__reference: OntologyPropertyReference = name
        super().__init__(tenant_id, name.iri, icon, labels, comments, context)

    @property
    def is_data_property(self) -> bool:
        return self.kind != PropertyType.OBJECT_PROPERTY

    @property
    def kind(self) -> PropertyType:
        """Kind of the property."""
        return self.__kind

    @property
    def reference(self) -> OntologyPropertyReference:
        """Reference to property"""
        return self.__reference

    @property
    def subproperty_of(self) -> OntologyPropertyReference:
        """Reference to the super property"""
        return self.__subproperty_of

    @property
    def inverse_property_of(self) -> OntologyPropertyReference:
        """Reference to the inverse property"""
        return self.__inverse_property_of

    @property
    def domain(self) -> OntologyClassReference:
        """Domain of the property."""
        return self.__domain

    @property
    def range(self) -> Optional[Union[OntologyClassReference, DataPropertyType]]:
        return self.__range

    def __repr__(self):
        return f'<OntologyProperty> - [name:= {self.iri} domain:={self.domain}, range:={self.range}, ' \
               f'subproperty_of:={self.subproperty_of}, type:={self.kind}]'

    @classmethod
    def from_dict(cls, property_dict: Dict[str, Any]):
        labels: List[Label] = [] if property_dict[LABELS_TAG] is None else \
            [Label.create_from_dict(la) for la in property_dict[LABELS_TAG]]
        comments: List[Comment] = [] if property_dict['comments'] is None else \
            [Comment.create_from_dict(co) for co in property_dict['comments']]
        return OntologyProperty(INVERSE_PROPERTY_TYPE[property_dict['kind']],
                                property_dict['tenantId'], property_dict['context'],
                                OntologyPropertyReference.parse(property_dict['name']),
                                property_dict['icon'],
                                OntologyClassReference.parse(property_dict['domain'])
                                if property_dict['domain'] is not None else None,
                                OntologyClassReference.parse(property_dict['range'])
                                if property_dict['range'] is not None else None,
                                labels, comments,
                                OntologyPropertyReference.parse(property_dict['subPropertyOf'])
                                if property_dict['subPropertyOf'] is not None else None,
                                OntologyPropertyReference.parse(property_dict['inverseOf'])
                                if property_dict['inverseOf'] is not None else None)

    @classmethod
    def new(cls, kind: PropertyType) -> 'OntologyProperty':
        return OntologyProperty(kind, '', '',
                                OntologyPropertyReference.parse('http://www.w3.org/2002/07/owl#topObjectProperty'))


class EntityProperty(abc.ABC):
    """
    EntityProperty
    --------------
    Abstract class for the different types of properties.
    """
    pass


class DataProperty(EntityProperty):
    """
    DataProperty
    ------------
    Data property for entities.

    Parameter
    ---------
    content: Any
        Content
    literal_type: LiteralProperty
        OntologyPropertyReference type
    language_code: str
        Language code
    data_type: str
        Data type
    """

    def __init__(self, content: Any, property_ref: OntologyPropertyReference,
                 language_code: LanguageCode = LanguageCode('en_US'), data_type: DataPropertyType = None):
        self.__content: Any = content
        self.__language_code: LanguageCode = language_code
        self.__type: OntologyPropertyReference = property_ref
        self.__data_type: Optional[DataPropertyType] = data_type

    @property
    def data_property_type(self) -> OntologyPropertyReference:
        """Ontology type."""
        return self.__type

    @property
    def data_type(self) -> Optional[DataPropertyType]:
        """Data type (optional)."""
        return self.__data_type

    @property
    def value(self) -> Any:
        """Content of the data property."""
        return self.__content

    @property
    def language_code(self) -> LanguageCode:
        """Language code of the content."""
        return self.__language_code

    @staticmethod
    def create_from_dict(data_property_struct: dict):
        if CONTENT_TAG not in data_property_struct or \
                LANGUAGE_CODE_TAG not in data_property_struct \
                and DATA_PROPERTY_TAG not in data_property_struct:
            raise ValueError("Dict is does not contain a data_property structure.")
        data_property_type: str = data_property_struct[DATA_PROPERTY_TAG]
        data_type: DataPropertyType = DataPropertyType.STRING
        if DATA_TYPE_TAG in data_property_struct and data_property_struct[DATA_TYPE_TAG] is not None:
            if data_property_struct[DATA_TYPE_TAG] not in INVERSE_DATA_PROPERTY_TYPE_MAPPING:
                raise ValueError(f"DataProperty data type is not supported. Type: {data_type}")
            else:
                data_type = INVERSE_DATA_PROPERTY_TYPE_MAPPING[data_property_struct[DATA_TYPE_TAG]]
        return DataProperty(data_property_struct[CONTENT_TAG],
                            OntologyPropertyReference.parse(data_property_type),
                            data_property_struct[LANGUAGE_CODE_TAG], data_type)

    def __dict__(self):
        return {
            CONTENT_TAG: self.value,
            LANGUAGE_CODE_TAG: self.language_code,
            DATA_PROPERTY_TAG: self.data_property_type.iri,
            DATA_TYPE_TAG: None if self.data_type is None else self.data_type.value
        }

    def __repr__(self):
        return f'{self.value}@{self.language_code}<{self.data_property_type.name}>'

    @staticmethod
    def create_from_list(param: List[dict]) -> List['DataProperty']:
        DataProperty.create_from_dict(param[0])
        return [DataProperty.create_from_dict(p) for p in param]


class ObjectProperty(EntityProperty):
    """
    Object Property
    ---------------
    ObjectProperty for entities.

    Parameter
    ---------
    relation: OntologyPropertyReference
        OntologyPropertyReference type
    incoming: List[str] (default:= [])
        Incoming relations
    outgoing: List[str] (default:= [])
        Outgoing relations
    """

    def __init__(self, relation: OntologyPropertyReference,
                 incoming: Optional[List[Union[str, 'ThingObject']]] = None,
                 outgoing: Optional[List[Union[str, 'ThingObject']]] = None):
        self.__relation: OntologyPropertyReference = relation
        self.__incoming: List[Union[str, 'ThingObject']] = incoming if incoming is not None else []
        self.__outgoing: List[Union[str, 'ThingObject']] = outgoing if outgoing is not None else []

    @property
    def relation(self) -> OntologyPropertyReference:
        """Reference from the ontology."""
        return self.__relation

    @property
    def incoming_relations(self) -> List[Union[str, 'ThingObject']]:
        """Incoming relation"""
        return self.__incoming

    @property
    def outgoing_relations(self) -> List[Union[str, 'ThingObject']]:
        return self.__outgoing

    @staticmethod
    def create_from_dict(relation_struct: Dict[str, Any]) -> Tuple[OntologyPropertyReference, 'ObjectProperty']:
        relation_type: OntologyPropertyReference = \
            OntologyPropertyReference.parse(relation_struct[RELATION_TAG])
        incoming: List[Union[str, ThingObject]] = []

        for incoming_relation in relation_struct[INCOMING_TAG]:
            if isinstance(incoming_relation, dict):
                incoming.append(ThingObject.from_dict(incoming_relation))
            elif isinstance(incoming_relation, str):
                incoming.append(incoming_relation)

        outgoing: List[Union[str, ThingObject]] = []
        for outgoing_relation in relation_struct[OUTGOING_TAG]:
            if isinstance(outgoing_relation, dict):
                outgoing.append(ThingObject.from_dict(outgoing_relation))
            elif isinstance(outgoing_relation, str):
                outgoing.append(outgoing_relation)
        return relation_type, ObjectProperty(relation_type, incoming, outgoing)

    def __dict__(self):
        return {
            RELATION_TAG: self.relation.iri,
            INCOMING_TAG: [e.uri if isinstance(e, ThingObject) else e for e in self.incoming_relations],
            OUTGOING_TAG: [e.uri if isinstance(e, ThingObject) else e for e in self.outgoing_relations]
        }

    def __repr__(self):
        return f'{self.relation.iri}, in:={self.incoming_relations}, out:={self.outgoing_relations}'

    @staticmethod
    def create_from_list(param: List[dict]) -> Dict[OntologyPropertyReference, 'ObjectProperty']:
        return dict([ObjectProperty.create_from_dict(p) for p in param])


class Ontology(object):
    """
    Ontology
    --------
    The ontology consists of classes and properties.
    """

    def __init__(self):
        self.__classes: Dict[str, OntologyClass] = {}
        self.__data_properties: Dict[str, OntologyProperty] = {}
        self.__object_properties: Dict[str, OntologyProperty] = {}

    def add_class(self, class_obj: OntologyClass):
        """
        Adding class object.

        Parameters
        ----------
        class_obj: OntologyClass
            Class object
        """
        self.__classes[class_obj.reference.iri] = class_obj

    def add_properties(self, prop_obj: OntologyProperty):
        """
        Adding properties.

        Parameters
        ----------
        prop_obj: OntologyProperty
        """
        if prop_obj.is_data_property:
            self.__data_properties[prop_obj.reference.iri] = prop_obj
        else:
            self.__object_properties[prop_obj.reference.iri] = prop_obj

    @property
    def data_properties(self) -> List[OntologyProperty]:
        """All data properties."""
        return list(self.__data_properties.values())

    @property
    def object_properties(self) -> List[OntologyProperty]:
        """All object properties."""
        return list(self.__object_properties.values())

    @property
    def classes(self) -> List[OntologyClass]:
        """All classes."""
        return list(self.__classes.values())

    def __check_hierarchy__(self, clz: OntologyClassReference, domain: OntologyClassReference) -> bool:
        current_clz: OntologyClass = self.get_class(clz)
        while current_clz is not None:
            if current_clz.reference == domain:
                return True
            current_clz = self.get_class(current_clz.subclass_of)
        return False

    def get_class(self, class_reference: OntologyClassReference) -> Optional[OntologyClass]:
        """
        Get class instance by reference.

        Parameters
        ----------
        class_reference: OntologyClassReference
            Class reference

        Returns
        --------
        instance: Optional[OntologyClass]
            Instance of ontology class.
        """
        return self.__classes.get(class_reference.iri)

    def get_object_properties(self, property_reference: OntologyPropertyReference) -> Optional[OntologyProperty]:
        """
        Get object property instance by reference.

        Parameters
        ----------
        property_reference: OntologyPropertyReference
            Property reference

        Returns
        --------
        instance: Optional[OntologyProperty]
            Instance of ontology object property.
        """
        return self.__object_properties.get(property_reference.iri)

    def get_data_properties(self, property_reference: OntologyPropertyReference) -> Optional[OntologyProperty]:
        """
        Get object property instance by reference.

        Parameters
        ----------
        property_reference: OntologyPropertyReference
            Property reference

        Returns
        --------
        instance: Optional[OntologyProperty]
            Instance of ontology object property.
        """
        return self.__data_properties.get(property_reference.iri)

    def data_properties_for(self, cls_reference: OntologyClassReference) -> List[OntologyPropertyReference]:
        """
        Retrieve a list of data properties.

        Parameters
        ----------
        cls_reference: OntologyClassReference
            Class of the ontology

        Returns
        -------
        data_properties: List[OntologyPropertyReference]
            List of data properties, where domain fit for the class of one of its super classes.
        """
        data_properties: List[OntologyPropertyReference] = []
        clz: Optional[OntologyClass] = self.get_class(cls_reference)
        if clz is not None:
            for dp in self.data_properties:
                if self.__check_hierarchy__(clz.reference, dp.domain):
                    data_properties.append(dp.reference)
        return data_properties

    def object_properties_for(self, cls_reference: OntologyClassReference) -> List[OntologyPropertyReference]:
        """
        Retrieve a list of object properties.

        Parameters
        ----------
        cls_reference: OntologyClassReference
            Class of the ontology

        Returns
        -------
        object_properties: List[OntologyPropertyReference]
            List of object properties, where domain fit for the class of one of its super classes.
        """
        object_properties: List[OntologyPropertyReference] = []
        clz: Optional[OntologyClass] = self.get_class(cls_reference)
        if clz is not None:
            for dp in self.object_properties:
                if self.__check_hierarchy__(clz.reference, dp.domain):
                    object_properties.append(dp.reference)
        return object_properties

    def __repr__(self):
        return f'<Ontology> : classes:= {self.classes}'


class ThingObject(abc.ABC):
    """
    ThingObject
    -----------
    Generic entity within knowledge graph.

    Each entity is derived from this object, thus all entity shares:
    - **uri**: A unique resource identity to identify the entity and reference it in relations
    - **label**: Human understandable label
    - **icon**: Visual representation of the entity
    - **description**: Description of entity
    - **concept_type**: Type of the concept
    - **concept_type_info**: Information on the concept type

    Parameters
    ----------
    label: List[Label]
        List of labels
    icon: str (optional)
        Icon
    description: List[Description] (optional)
        List of descriptions
    concept_type: OntologyClassReference
        Type of the concept
    uri: str
         URI for entity. For new entities the URI is None, as the knowledge graph backend assigns this.
    tenant_rights: TenantAccessRight
        Rights for tenants
    owner: bool
        Is the logged in user the owner of the entity
    """

    def __init__(self, label: List[Label] = None, concept_type: OntologyClassReference = THING_CLASS,
                 description: Optional[List[Description]] = None, uri: Optional[str] = None, icon: Optional[str] = None,
                 tenant_rights: TenantAccessRight = TenantAccessRight(), owner: bool = True):
        self.__uri: str = uri
        self.__icon: Optional[str] = icon
        self.__label: List[Label] = label if label else []
        self.__description: List[Description] = description if description else []
        self.__alias: List[Label] = []
        self.__concept_type: OntologyClassReference = concept_type
        self.__data_properties: Dict[OntologyPropertyReference, List[DataProperty]] = {}
        self.__object_properties: Dict[OntologyPropertyReference, ObjectProperty] = {}
        self.__tenants_rights: TenantAccessRight = tenant_rights
        self.__status_flag: EntityStatus = EntityStatus.UNKNOWN
        self.__ontology_types: Optional[Set[str]] = None
        self.__owner: bool = owner

    @property
    def uri(self) -> str:
        """Unique identifier for entity."""
        return self.__uri

    @uri.setter
    def uri(self, uri: str):
        self.__uri = uri

    @property
    def owner(self) -> bool:
        """Is current user the owner of the entity."""
        return self.__owner

    @property
    def status_flag(self) -> EntityStatus:
        """Status flag."""
        return self.__status_flag

    @status_flag.setter
    def status_flag(self, flag: EntityStatus):
        self.__status_flag = flag

    @property
    def label(self) -> List[Label]:
        """Labels of the entity."""
        return self.__label

    @label.setter
    def label(self, value: List[Label]):
        self.__label = value

    def add_label(self, label: str, language_code: LanguageCode):
        """Adding an label for entity.

        Parameters
        ----------
        label: str
            Label
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        """
        self.__label.append(Label(label, language_code, True))

    def update_label(self, value: str, language_code: LanguageCode):
        """Update or creates a label for a specific language.

        Parameters
        ----------
        value: str
            Value to be set
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        """
        for label in self.label:
            if label.language_code == language_code:
                label.content = value
                return
        # Label with language does not exist, so create a new label
        self.add_label(value, language_code)

    def remove_label(self, language_code: LanguageCode):
        """
        Remove label for entity if it exists for language.

        Parameters
        ----------
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        """
        for idx in range(len(self.label)):
            if self.label[idx].language_code == language_code:
                del self.__label[idx]
                break

    def label_lang(self, language_code: LanguageCode) -> Optional[Label]:
        """
        Get label for language_code code.

        Parameters
        ----------
        language_code: LanguageCode
            Requested language_code code
        Returns
        -------
        label: Optional[Label]
            Returns the label for a specific language code
        """
        for label in self.label:
            if label.language_code == language_code:
                return label
        return None

    @property
    def source_system(self) -> Optional[List[DataProperty]]:
        """Source of the entity."""
        if SYSTEM_SOURCE_SYSTEM in self.__data_properties:
            return self.__data_properties[SYSTEM_SOURCE_SYSTEM]
        return None

    def add_source_system(self, value: DataProperty):
        """
        Adding the source system  of the entity.

        Parameters
        -----------
        value: DataProperty
            Adds the source system as a Data Property. **Remark:** The data property must have the property type
            'wacom:core#sourceSystem'.
        """
        if value.data_property_type != SYSTEM_SOURCE_SYSTEM:
            raise ValueError(f'Data property {value.data_property_type.iri} not supported. '
                             f'Expected:={SYSTEM_SOURCE_SYSTEM.iri}')
        if SYSTEM_SOURCE_SYSTEM not in self.__data_properties:
            self.__data_properties[SYSTEM_SOURCE_SYSTEM] = []
        for idx in range(0, len(self.__data_properties[SYSTEM_SOURCE_SYSTEM])):
            if self.__data_properties[SYSTEM_SOURCE_SYSTEM][idx].language_code == value.language_code:
                del self.__data_properties[SYSTEM_SOURCE_SYSTEM][idx]
        self.__data_properties[SYSTEM_SOURCE_SYSTEM].append(value)

    @property
    def source_reference_id(self) -> Optional[List[DataProperty]]:
        """Reference id for to the source."""
        if SYSTEM_SOURCE_REFERENCE_ID in self.__data_properties:
            return self.__data_properties[SYSTEM_SOURCE_REFERENCE_ID]
        return None

    def add_source_reference_id(self, value: DataProperty):
        """
        Adding the reference id from the source system of the entity.

        Parameters
        -----------
        value: DataProperty
            Adds the source system reference id as a Data Property.
            **Remark:** The data property must have the property type 'wacom:core#sourceReferenceId'.
        """
        if value.data_property_type != SYSTEM_SOURCE_REFERENCE_ID:
            raise ValueError(f'Data property {value.data_property_type.iri} not supported. '
                             f'Expected:={SYSTEM_SOURCE_REFERENCE_ID.iri}')
        if SYSTEM_SOURCE_REFERENCE_ID not in self.__data_properties:
            self.__data_properties[SYSTEM_SOURCE_REFERENCE_ID] = []
        for idx in range(0, len(self.__data_properties[SYSTEM_SOURCE_REFERENCE_ID])):
            if self.__data_properties[SYSTEM_SOURCE_REFERENCE_ID][idx].language_code == value.language_code:
                del self.__data_properties[SYSTEM_SOURCE_REFERENCE_ID][idx]
        self.__data_properties[SYSTEM_SOURCE_REFERENCE_ID].append(value)

    def default_source_reference_id(self, language_code: LanguageCode = LanguageCode('en_US')) -> Optional[str]:
        """
        Getting the source reference id for a certain language code.

        Parameters
        ----------
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.

        Returns
        -------
        id: str
            Source reference id.
        """
        if SYSTEM_SOURCE_REFERENCE_ID in self.__data_properties:
            for sr in self.data_properties[SYSTEM_SOURCE_REFERENCE_ID]:
                if sr.language_code == language_code:
                    return sr.value
        return None

    @property
    def image(self) -> Optional[str]:
        """Image depicting the entities (optional)."""
        return self.__icon

    @image.setter
    def image(self, value: str):
        self.__icon = value

    @property
    def description(self) -> Optional[List[Description]]:
        """Description of the thing (optional)."""
        return self.__description

    @description.setter
    def description(self, value: List[Description]):
        self.__description = value

    def add_description(self, description: str, language_code: LanguageCode):
        """Adding an description for entity.

        Parameters
        ----------
        description: str
            Description
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        """
        self.__description.append(Description(description=description, language_code=language_code))

    def update_description(self, value: str, language_code: LanguageCode):
        """Update or creates a description for a specific language.

        Parameters
        ----------
        value: str
            Value to be set
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        """
        for desc in self.description:
            if desc.language_code == language_code:
                desc.content = value
                return
        # Description with language does not exist, so create a new description
        self.add_description(value, language_code)

    def description_lang(self, language_code: str) -> Optional[Description]:
        """
        Get description for entity.

        Parameters
        ----------
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        Returns
        -------
        label: LocalizedContent
            Returns the  label for a specific language_code code
        """
        for desc in self.description:
            if desc.language_code == language_code:
                return desc
        return None

    def remove_description(self, language_code: LanguageCode):
        """
        Remove description for entity if it exists for language.

        Parameters
        ----------
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        """
        for idx in range(len(self.description)):
            if self.description[idx].language_code == language_code:
                del self.__description[idx]
                break

    @property
    def concept_type(self) -> OntologyClassReference:
        """Concept type."""
        return self.__concept_type

    @concept_type.setter
    def concept_type(self, value: OntologyClassReference):
        self.__concept_type = value

    @property
    def ontology_types(self) -> Set[str]:
        """Ontology types. For public entities."""
        return self.__ontology_types

    @ontology_types.setter
    def ontology_types(self, value: Set[str]):
        self.__ontology_types = value

    @property
    def data_properties(self) -> Dict[OntologyPropertyReference, List[DataProperty]]:
        """Literals of the concept."""
        return self.__data_properties

    @data_properties.setter
    def data_properties(self, data_properties: Dict[OntologyPropertyReference, List[DataProperty]]):
        """Literals of the concept."""
        self.__data_properties = data_properties

    @property
    def object_properties(self) -> Dict[OntologyPropertyReference, ObjectProperty]:
        """Relations of the concept."""
        return self.__object_properties

    @object_properties.setter
    def object_properties(self, relations: Dict[OntologyPropertyReference, ObjectProperty]):
        self.__object_properties = relations

    @property
    def alias(self) -> List[Label]:
        """Alternative labels of the concept."""
        return self.__alias

    @alias.setter
    def alias(self, alias: List[Label]):
        self.__alias = alias

    def alias_lang(self, language_code: LanguageCode) -> List[Label]:
        """
        Get alias for language_code code.

        Parameters
        ----------
        language_code: LanguageCode
            Requested language_code code
        Returns
        -------
        aliases: List[Label]
            Returns a list of aliases for a specific language code
        """
        aliases: List[Label] = []
        for a in self.alias:
            if a.language_code == language_code:
                aliases.append(a)
        return aliases

    def update_alias(self, value: str, language_code: LanguageCode):
        """Update or creates a alias for a specific language.

        Parameters
        ----------
        value: str
            Value to be set
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        """
        for a in self.alias:
            if a.language_code == language_code:
                a.content = value
                return
        # Label with language does not exist, so create a new label
        self.add_alias(value, language_code=language_code)

    def add_relation(self, prop: ObjectProperty):
        """Adding a relation to the entity.

        :param prop: ObjectProperty -
            ObjectProperty
        """
        if prop.relation in self.object_properties:
            self.__object_properties[prop.relation].incoming_relations.extend(prop.incoming_relations)
            self.__object_properties[prop.relation].outgoing_relations.extend(prop.outgoing_relations)
        else:
            self.__object_properties[prop.relation] = prop

    def add_data_property(self, data_property: DataProperty):
        """Add data property to the entity.

        Parameters
        ----------
        data_property: DataProperty
            DataProperty
        """
        if data_property.data_property_type not in self.__data_properties:
            self.__data_properties[data_property.data_property_type] = []
        self.__data_properties[data_property.data_property_type].append(data_property)

    def add_alias(self, alias: str, language_code: LanguageCode):
        """Adding an alias for entity.

        Parameters
        ----------
        alias: str
            Alias
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        """
        self.__alias.append(Label(alias, language_code, False))

    @property
    def tenant_access_right(self) -> TenantAccessRight:
        """Access rights for tenant. """
        return self.__tenants_rights

    @tenant_access_right.setter
    def tenant_access_right(self, rights: TenantAccessRight):
        self.__tenants_rights = rights

    def __dict__(self):
        labels: List[Dict[str, Any]] = []
        labels.extend([la.__dict__() for la in self.label])
        labels.extend([la.__dict__() for la in self.alias])
        dict_object: Dict[str, Any] = {
            URI_TAG: self.uri,
            IMAGE_TAG: self.image,
            LABELS_TAG: labels,
            DESCRIPTIONS_TAG: [desc.__dict__() for desc in self.description],
            TYPE_TAG: self.concept_type.iri,
            STATUS_FLAG_TAG: self.status_flag.value,
            DATA_PROPERTIES_TAG: {},
            OBJECT_PROPERTIES_TAG: {},
            OWNER_TAG: self.owner
        }
        for literal_type, items in self.data_properties.items():
            dict_object[DATA_PROPERTIES_TAG][literal_type.iri] = [i.__dict__() for i in items]
        for relation_type, item in self.object_properties.items():
            dict_object[OBJECT_PROPERTIES_TAG][relation_type.iri] = item.__dict__()

        return dict_object

    @staticmethod
    def from_dict(entity: Dict[str, Any]) -> 'ThingObject':
        labels: List[Label] = []
        alias: List[Label] = []
        descriptions: List[Description] = []

        for label in entity[LABELS_TAG]:
            if label[IS_MAIN_TAG]:
                labels.append(Label.create_from_dict(label))
            else:
                alias.append(Label.create_from_dict(label))

        for desc in entity[DESCRIPTIONS_TAG]:
            descriptions.append(Description.create_from_dict(desc))

        thing: ThingObject = ThingObject(label=labels, icon=entity[IMAGE_TAG], description=descriptions,
                                         uri=entity[URI_TAG],
                                         concept_type=OntologyClassReference.parse(entity[TYPE_TAG]),
                                         owner=entity.get(OWNER_TAG, True))
        if DATA_PROPERTIES_TAG in entity:
            if isinstance(entity[DATA_PROPERTIES_TAG], dict):
                for data_property_type_str, data_properties in entity[DATA_PROPERTIES_TAG].items():
                    data_property_type: OntologyPropertyReference = \
                        OntologyPropertyReference.parse(data_property_type_str)
                    for data_property in data_properties:
                        language_code: LanguageCode = LanguageCode(data_property[LANGUAGE_CODE_TAG])
                        value: str = data_property[VALUE_TAG]
                        thing.add_data_property(DataProperty(value, data_property_type, language_code))
            elif isinstance(entity[DATA_PROPERTIES_TAG], list):
                for data_property in entity[DATA_PROPERTIES_TAG]:
                    language_code: LanguageCode = LanguageCode(data_property[LANGUAGE_CODE_TAG])
                    value: str = data_property[VALUE_TAG]
                    data_property_type: OntologyPropertyReference = \
                        OntologyPropertyReference.parse(data_property[DATA_PROPERTY_TAG])
                    thing.add_data_property(DataProperty(value, data_property_type, language_code))
        if OBJECT_PROPERTIES_TAG in entity:
            for object_property in entity[OBJECT_PROPERTIES_TAG].values():
                prop, obj = ObjectProperty.create_from_dict(object_property)
                thing.add_relation(obj)
        thing.alias = alias
        # Finally, retrieve rights
        if TENANT_RIGHTS_TAG in entity:
            thing.tenant_access_right = TenantAccessRight.parse(entity[TENANT_RIGHTS_TAG])
        return thing

    def __hash__(self):
        return 0

    def __eq__(self, other):
        # another object is equal to self, iff
        # it is an instance of MyClass
        return isinstance(other, ThingObject) and other.uri == self.uri

    def __repr__(self):
        return f'<{self.concept_type.iri if self.__concept_type else "UNSET"}: uri:={self.uri}, labels:={self.label}, '\
               f'tenant access right:={self.tenant_access_right}]>'


def update_language_code(lang: str):
    return language_code_mapping.get(lang, lang)


def localized_list_description(entity_dict: Dict[str, str]) -> List[Description]:
    return [Description(cont, update_language_code(lang)) for lang, cont in entity_dict.items()]


def localized_list_label(entity_dict: Dict[str, str]) -> List[Label]:
    return [Label(cont, update_language_code(lang), main=True) for lang, cont in entity_dict.items()]


def localized_flatten_alias_list(entity_dict: Dict[str, List[str]]) -> List[Label]:
    flatten: List[Label] = []
    for language, items in entity_dict.items():
        for i in items:
            flatten.append(Label(i, update_language_code(language), main=False))
    return flatten


def from_dict(entity: Dict[str, Any], concept_type: OntologyClassReference) -> 'ThingObject':
    labels: List[Label] = localized_list_label(entity['label'])
    description: List[Description] = localized_list_description(entity['description'])
    alias: List[Label] = localized_flatten_alias_list(entity['alias'])
    if IMAGE_TAG in entity:
        icon: str = entity[IMAGE_TAG]
    else:
        logger.warning(f"Entity has no image: {entity}")
        icon: str = ''
    # Create the entity
    thing: ThingObject = ThingObject(label=labels, concept_type=concept_type, description=description, icon=icon)
    thing.alias = alias
    if STATUS_FLAG_TAG in entity:
        thing.status_flag = entity[STATUS_FLAG_TAG]
    return thing


# -------------------------------------------------- Encoder -----------------------------------------------------------
class ThingEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Label):
            return o.__dict__()
        elif isinstance(o, Description):
            return o.__dict__()
        elif isinstance(o, ThingObject):
            return o.__dict__()
        return str(o)



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.
from typing import List, Dict

import requests
from requests import Response

from knowledge.services.base import WacomServiceAPIClient, WacomServiceException
from knowledge.services.graph import AUTHORIZATION_HEADER_FLAG


class TenantManagementServiceAPI(WacomServiceAPIClient):
    """
    Tenant Management Service API
    -----------------------------

    Functionality:
        - List all tenants
        - Create tenants
        - Create users

    Parameters
    ----------
    tenant_token: str
        Tenant Management token
    service_url: str
        URL of the service
    service_endpoint: str
        Base endpoint
    """

    TENANT_ENDPOINT: str = 'tenant'
    USER_DETAILS_ENDPOINT: str = f'{WacomServiceAPIClient.USER_ENDPOINT}/users'
    SERVICE_URL: str = 'https://semantic-ink-private.wacom.com'

    def __init__(self, tenant_token: str, service_url: str = SERVICE_URL, service_endpoint: str = 'graphdata'):
        self.__tenant_management_token: str = tenant_token
        super().__init__("TenantManagementServiceAPI", service_url=service_url, service_endpoint=service_endpoint)

    @property
    def tenant_management_token(self) -> str:
        """Tenant Management token."""
        return self.__tenant_management_token

    @tenant_management_token.setter
    def tenant_management_token(self, value: str):
        self.__tenant_management_token = value

    # ------------------------------------------ Tenants handling ------------------------------------------------------

    def create_tenant(self, name: str) -> Dict[str, str]:
        """
        Creates a tenant.

        Parameters
        ----------
        name: str -
            Name of the tenant

        Returns
        -------
        tenant_dict: Dict[str, str]

        Newly created tenant structure.
        >>>     {
        >>>       "id": "<Tenant-ID>",
        >>>       "apiKey": "<Tenant-API-Key>",
        >>>       "name": "<Tenant-Name>"
        >>>    }

        Raises
        ------
        WacomServiceException
            If the tenant service returns an error code.
        """
        url: str = '{}/{}{}'.format(self.service_url, self.service_endpoint,
                                    TenantManagementServiceAPI.TENANT_ENDPOINT)
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {self.__tenant_management_token}',
            'Content-Type': 'application/json'
        }
        payload: dict = {
            'name': name
        }
        response: Response = requests.post(url, headers=headers, json=payload, verify=self.verify_calls)
        if response.ok:
            return response.json()

    def listing_tenant(self) -> List[Dict[str, str]]:
        """
        Listing all tenants configured for this instance.

        Returns
        -------
        tenants:  List[Dict[str, str]]
            List of tenants:
            >>> [
            >>>     {
            >>>        "id": "<Tenant-ID>",
            >>>        "apiKey": "<Tenant-API-Key>",
            >>>        "name": "<Tenant-Name>"
            >>>     },
            >>>     ...
            >>> ]
        Raises
        ------
        WacomServiceException
            If the tenant service returns an error code.
        """
        url: str = '{}/{}{}'.format(self.service_url, self.service_endpoint,
                                    TenantManagementServiceAPI.TENANT_ENDPOINT)
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {self.__tenant_management_token}'
        }
        response: Response = requests.get(url, headers=headers, data={}, verify=self.verify_calls)
        if response.ok:
            return response.json()
        raise WacomServiceException(f'Response code:={response.status_code}, exception:= {response.text}')




# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.
from typing import List, Optional, Dict

import requests
from requests import Response

from knowledge.base.entity import LANGUAGE_CODE_TAG, LanguageCode, TEXT_TAG
from knowledge.base.ontology import OntologyClassReference
from knowledge.nel.base import PersonalEntityLinkingProcessor, EntitySource, KnowledgeSource, \
    KnowledgeGraphEntity, EntityType
from knowledge.services.base import WacomServiceException
from knowledge.services.graph import AUTHORIZATION_HEADER_FLAG, CONTENT_TYPE_HEADER_FLAG


class WacomEntityLinkingEngine(PersonalEntityLinkingProcessor):
    """
    Wacom Engine
    ------------
    Performing Wacom's Named entity linking.

    Parameter
    ---------
    service_url: str
        URL of the service
    """
    SERVICE_ENDPOINT: str = 'graphdata/nel/text'
    SERVICE_URL: str = 'https://semantic-ink-private.wacom.com'
    LANGUAGES: List[LanguageCode] = [
        LanguageCode('de_DE'),
        LanguageCode('en_US'),
        LanguageCode('en_GB')
    ]

    def __init__(self, service_url: str = SERVICE_URL, service_endpoint: str = SERVICE_ENDPOINT):
        self.__service_endpoint: str = service_endpoint
        super().__init__(supported_languages=WacomEntityLinkingEngine.LANGUAGES, service_url=service_url)

    def link_personal_entities(self, auth_key: str, text: str, language_code: str = 'en_US') \
            -> List[KnowledgeGraphEntity]:
        """
        Performs Named Entity Linking on a text. It only finds entities which are accessible by the user identified by
        the auth key.

        Parameters
        ----------
        auth_key: str
            Auth key identifying a user within the Wacom personal knowledge service.
        text: str
            Text where the entities shall be tagged in.
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.

        Returns
        -------
        entities: List[KnowledgeGraphEntity]
            List of knowledge graph entities.

        Raises
        ------
        WacomServiceException
            If the Named Entity Linking service returns an error code.
        """
        named_entities: List[KnowledgeGraphEntity] = []
        url: str = f'{self.service_url}/{self.__service_endpoint}'
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}',
            CONTENT_TYPE_HEADER_FLAG: 'application/json'
        }
        payload: Dict[str, str] = {
            LANGUAGE_CODE_TAG: language_code,
            TEXT_TAG: text
        }
        response: Response = requests.post(url, headers=headers, json=payload, verify=self.verify_calls)
        if response.ok:
            results: dict = response.json()
            for e in results:
                entity_types: List[str] = []
                # --------------------------- Entity content -------------------------------------------------------
                source: Optional[EntitySource] = None
                if 'uri' in e:
                    source = EntitySource(e['uri'], KnowledgeSource.WACOM_KNOWLEDGE)
                # --------------------------- Ontology types -------------------------------------------------------
                if 'type' in e:
                    entity_types.append(e['type'])
                # --------------------------------------------------------------------------------------------------
                ne: KnowledgeGraphEntity = KnowledgeGraphEntity(ref_text=text[e['startPosition']:e['endPosition'] + 1],
                                                                start_idx=e['startPosition'], end_idx=e['endPosition'],
                                                                label=e['value'], confidence=0.,
                                                                source=source, content_link='',
                                                                ontology_types=entity_types,
                                                                entity_type=EntityType.PERSONAL_ENTITY)
                ne.relevant_type = OntologyClassReference.parse(e['type'])
                named_entities.append(ne)
            return named_entities
        raise WacomServiceException(f'Named entity linking for text:={text}@{language_code}. '
                                    f'Response code:={response.status_code}, exception:= {response.content}')



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom Authors. All Rights Reserved.
from typing import Optional

from rdflib import Graph, RDF, RDFS, OWL

from knowledge.base.ontology import Ontology, OntologyClass, OntologyClassReference, OntologyPropertyReference, \
    OntologyProperty, PropertyType, INVERSE_DATA_PROPERTY_TYPE_MAPPING, DataPropertyType


def ontology_import(rdf_content: str, tenant_id: str = '', context: str = '') -> Ontology:
    """Import Ontology from RDF ontology file.

    Parameters
    ----------
    rdf_content: str
        Content of the RDF content file.
    tenant_id: str (default:= '')
        Tenant ID.
    context: str (default:= '')
        Context file.

    Returns
    -------
    ontology: Ontology
        Instance of ontology.
    """
    rdf_graph: Graph = Graph().parse(data=rdf_content, format='xml')
    ontology: Ontology = Ontology()
    # Parse classes
    for cls_iri in [s for s, p, o in rdf_graph.triples((None, RDF.type, RDFS.Class))]:
        subclass_of: Optional[OntologyClassReference] = None
        for _, _, o in rdf_graph.triples((cls_iri, RDFS.subClassOf, None)):
            subclass_of = OntologyClassReference.parse(o)
        ontology.add_class(OntologyClass(tenant_id=tenant_id, context=context,
                                         reference=OntologyClassReference.parse(cls_iri),
                                         subclass_of=subclass_of))
    # Parse data properties
    for data_property_iri in [s for s, p, o in rdf_graph.triples((None, RDF.type, OWL.DatatypeProperty))]:
        subproperty_of: Optional[OntologyPropertyReference] = None
        range_prop: Optional[DataPropertyType] = None
        domain_prop: Optional[OntologyClassReference] = None
        inverse_prop: Optional[OntologyPropertyReference] = None
        for _, _, o in rdf_graph.triples((data_property_iri, RDFS.range, None)):
            range_prop = INVERSE_DATA_PROPERTY_TYPE_MAPPING[str(o)]
        for _, _, o in rdf_graph.triples((data_property_iri, RDFS.domain, None)):
            domain_prop = OntologyClassReference.parse(o)
        for _, _, o in rdf_graph.triples((data_property_iri, OWL.inverseOf, None)):
            inverse_prop = OntologyPropertyReference.parse(o)
        for _, _, o in rdf_graph.triples((data_property_iri, RDFS.subPropertyOf, None)):
            subproperty_of = OntologyPropertyReference.parse(o)
        ontology.add_properties(OntologyProperty(kind=PropertyType.DATA_PROPERTY, tenant_id=tenant_id, context=context,
                                                 name=OntologyPropertyReference.parse(data_property_iri),
                                                 property_range=range_prop, property_domain=domain_prop,
                                                 subproperty_of=subproperty_of, inverse_property_of=inverse_prop))
    # Parse object properties
    for object_property_iri in [s for s, p, o in rdf_graph.triples((None, RDF.type, OWL.ObjectProperty))]:
        subproperty_of: Optional[OntologyPropertyReference] = None
        range_prop: Optional[OntologyPropertyReference] = None
        domain_prop: Optional[OntologyPropertyReference] = None
        inverse_prop: Optional[OntologyPropertyReference] = None
        for _, _, o in rdf_graph.triples((object_property_iri, RDFS.range, None)):
            range_prop = OntologyClassReference.parse(o)
        for _, _, o in rdf_graph.triples((object_property_iri, RDFS.domain, None)):
            domain_prop = OntologyClassReference.parse(o)
        for _, _, o in rdf_graph.triples((object_property_iri, OWL.inverseOf, None)):
            inverse_prop = OntologyPropertyReference.parse(o)
        for _, _, o in rdf_graph.triples((object_property_iri, RDFS.subPropertyOf, None)):
            subproperty_of = OntologyPropertyReference.parse(o)
        ontology.add_properties(OntologyProperty(kind=PropertyType.OBJECT_PROPERTY, tenant_id=tenant_id,
                                                 context=context,
                                                 name=OntologyPropertyReference.parse(object_property_iri),
                                                 property_range=range_prop, property_domain=domain_prop,
                                                 subproperty_of=subproperty_of, inverse_property_of=inverse_prop))
    return ontology



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.
from typing import List


class AccessRight(object):
    """
    Access rights for entities within a tenant.

    Parameters
    ----------
    read: bool (default := False)
        Read access for entity within tenant.
    write: bool (default := False)
        Write access for entity within tenant.
    delete: bool (default := False)
        Delete access for entity within tenant.
    """
    READ: str = "Read"
    WRITE: str = "Write"
    DELETE: str = "Delete"

    def __init__(self, read: bool = False, write: bool = False, delete: bool = False):
        self.__read: bool = read
        self.__write: bool = write
        self.__delete: bool = delete

    @property
    def read(self) -> bool:
        """Read access for tenant."""
        return self.__read

    @read.setter
    def read(self, value: bool):
        self.__read = value

    @property
    def write(self) -> bool:
        """Write access for tenant."""
        return self.__write

    @write.setter
    def write(self, value: bool):
        self.__write = value

    @property
    def delete(self) -> bool:
        """Delete access for tenant."""
        return self.__delete

    @delete.setter
    def delete(self, value: bool):
        self.__delete = value

    def __repr__(self):
        result: str = '['
        prefix: str = ''
        if self.read:
            result += TenantAccessRight.READ
            prefix = ", "
        if self.write:
            result += prefix + TenantAccessRight.WRITE
        if self.delete:
            result += TenantAccessRight.DELETE
        result += ']'
        return result

    @classmethod
    def parse(cls, param: List[str]) -> 'TenantAccessRight':
        """
        Converts the access to list of properties.

        Parameters
        ----------
        param: List[str]
            List of rights

        Returns
        -------
        tenant_rights: TenantAccessRight
            Instantiated rights.
        """
        rights: TenantAccessRight = TenantAccessRight()
        rights.read = TenantAccessRight.READ in param
        rights.write = TenantAccessRight.WRITE in param
        rights.delete = TenantAccessRight.DELETE in param
        return rights

    def to_list(self) -> List[str]:
        """
        Converts the access to list of properties.

        Returns
        -------
        access_list: List[str]
            List of rights
        """
        rights: List[str] = []
        if self.read:
            rights.append(TenantAccessRight.READ)
        if self.write:
            rights.append(TenantAccessRight.WRITE)
        if self.delete:
            rights.append(TenantAccessRight.DELETE)
        return rights


class TenantAccessRight(AccessRight):
    """
    TenantAccessRight
    -----------------
    Access rights for entities within a tenant.

    Parameters
    ----------
    read: bool (default := False)
        Read access for entity within tenant.
    write: bool (default := False)
        Write access for entity within tenant.
    delete: bool (default := False)
        Delete access for entity within tenant.
    """

    def __init__(self, read: bool = False, write: bool = False, delete: bool = False):
        super().__init__(read, write, delete)


class GroupAccessRight(AccessRight):
    """
    GroupAccessRight
    -----------------
    Group rights for entities within a group.

    Parameters
    ----------
    read: bool (default := False)
        Read access for entity within tenant.
    write: bool (default := False)
        Write access for entity within tenant.
    delete: bool (default := False)
        Delete access for entity within tenant.
    """

    def __init__(self, read: bool = False, write: bool = False, delete: bool = False):
        super().__init__(read, write, delete)



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.
import enum
import urllib.parse
from typing import List, Dict, Any, Tuple, Optional

import requests
from requests import Response

from knowledge.base.access import TenantAccessRight
from knowledge.base.entity import LanguageCode, DATA_PROPERTIES_TAG, DATA_PROPERTY_TAG, \
    LANGUAGE_CODE_TAG, VALUE_TAG, IMAGE_TAG, DESCRIPTION_TAG, TYPE_TAG, URI_TAG, LABELS_TAG, IS_MAIN_TAG, \
    DESCRIPTIONS_TAG, RELATIONS_TAG, EntityStatus, Label, Description
from knowledge.base.ontology import DataProperty, OntologyPropertyReference, ThingObject, OntologyClassReference, \
    ObjectProperty
from knowledge.services.base import WacomServiceAPIClient, WacomServiceException

# ------------------------------------------------- Constants ----------------------------------------------------------
SEARCH_TERM: str = 'searchTerm'
LANGUAGE_PARAMETER: str = 'language'
TYPES_PARAMETER: str = 'types'
LIMIT_PARAMETER: str = 'limit'
AUTHORIZATION_HEADER_FLAG: str = 'Authorization'
CONTENT_TYPE_HEADER_FLAG: str = 'Content-Type'
LISTING: str = 'listing'
TOTAL_COUNT: str = 'totalCount'
TARGET: str = 'target'
OBJECT: str = 'object'
PREDICATE: str = 'predicate'
SUBJECT: str = 'subject'
LIMIT: str = 'limit'
OBJECT_URI: str = 'objectUri'
RELATION_URI: str = 'relationUri'
SUBJECT_URI: str = 'subjectUri'
NEXT_PAGE_ID_TAG: str = 'nextPageId'
TENANT_RIGHTS_TAG: str = 'tenantRights'
RELATION_TAG: str = 'relation'


# ------------------------------- Enum ---------------------------------------------------------------------------------
class SearchPattern(enum.Enum):
    """
    SearchPattern
    -------------
    Different search pattern for literal search.
    """
    REGEX = 'regex'
    GT = 'gt'
    GTE = 'gte'
    LT = 'lt'
    LTE = 'lte'
    EQ = 'eq'
    RANGE = 'range'


# -------------------------------------------- Service API Client ------------------------------------------------------
class WacomKnowledgeService(WacomServiceAPIClient):
    """
    WacomKnowledgeService
    ---------------------
    Client for the Semantic Ink Privat knowledge system.

    Operations for concepts:
        - Creation of concepts
        - Update of concepts
        - Deletion of concepts
        - Listing of concepts

    Parameters
    ----------
    application_name: str
        Name of the application using the service
    service_url: str
        URL of the service
    service_endpoint: str
        Base endpoint
    """
    PERSONAL_KNOWLEDGE_URL: str = 'https://semantic-ink-private.wacom.com'
    USER_ENDPOINT: str = 'user'
    CONCEPT_ENDPOINT: str = 'entity'
    ACTIVATIONS_ENDPOINT: str = 'entity/activations'
    DELETION_ENDPOINT: str = 'entity/{}'
    LABELS_ENDPOINT: str = 'entity/{}/labels'
    LISTING_ENDPOINT: str = 'entity/types'
    LITERAL_ENDPOINT: str = 'entity/{}/literals'
    RELATION_ENDPOINT: str = 'entity/{}/relation'
    RELATIONS_ENDPOINT: str = 'entity/{}/relations'
    SEARCH_LABELS_ENDPOINT: str = "semantic-search/labels"
    SEARCH_TYPES_ENDPOINT: str = "semantic-search/types"
    SEARCH_LITERALS_ENDPOINT: str = "semantic-search/literal"
    SEARCH_DESCRIPTION_ENDPOINT: str = "semantic-search/description"
    SEARCH_RELATION_ENDPOINT: str = "semantic-search/relation"
    ONTOLOGY_UPDATE_ENDPOINT: str = 'ontology-update'

    def __init__(self, application_name: str, service_url: str = PERSONAL_KNOWLEDGE_URL,
                 service_endpoint: str = 'graphdata'):
        super().__init__(application_name, service_url, service_endpoint)

    def entity(self, auth_key: str, uri: str) -> ThingObject:
        """
        Retrieve entity information from personal knowledge, using the  URI as identifier.

        **Remark:** Object properties (relations) must be requested separately.

        Parameters
        ----------
        auth_key: str
            Auth key identifying a user within the Wacom personal knowledge service.
        uri: str
            URI of concept

        Returns
        -------
        thing: ThingObject
            Entity with is type URI, description, an image/icon, and tags (labels).

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code or the entity is not found in the knowledge graph
        """
        url: str = f'{self.service_base_url}{WacomKnowledgeService.CONCEPT_ENDPOINT}/{urllib.parse.quote(uri)}'
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        response: Response = requests.get(url, headers=headers, verify=self.verify_calls)
        if response.ok:
            e: Dict[str, Any] = response.json()
            pref_label: List[Label] = []
            aliases: List[Label] = []
            # Extract labels and alias
            for label in e[LABELS_TAG]:
                if label[IS_MAIN_TAG]:  # Labels
                    pref_label.append(Label.create_from_dict(label))
                else:  # Alias
                    aliases.append(Label.create_from_dict(label))
            # Create ThingObject
            thing: ThingObject = ThingObject(label=pref_label, icon=e[IMAGE_TAG],
                                             description=[Description.create_from_dict(d) for d in e[DESCRIPTIONS_TAG]],
                                             concept_type=OntologyClassReference.parse(e[TYPE_TAG]),
                                             uri=e[URI_TAG])
            # Set the alias
            thing.alias = aliases
            # Configure data properties
            if DATA_PROPERTIES_TAG in e:
                for data_property in e[DATA_PROPERTIES_TAG]:
                    data_property_type: OntologyPropertyReference = \
                        OntologyPropertyReference.parse(data_property[DATA_PROPERTY_TAG])
                    language_code: LanguageCode = data_property[LANGUAGE_CODE_TAG]
                    value: str = data_property[VALUE_TAG]
                    thing.add_data_property(DataProperty(value, data_property_type, language_code))
            # Tenant rights
            if TENANT_RIGHTS_TAG in e:
                thing.tenant_access_right = TenantAccessRight.parse(e[TENANT_RIGHTS_TAG])
            else:
                thing.tenant_access_right = TenantAccessRight()
            return thing
        raise WacomServiceException(f'Pushing concept failed. '
                                    f'Response code:={response.status_code}, exception:= {response.content}')

    def delete_entities(self, auth_key: str, uris: List[str], force: bool = False):
        """
        Delete a list of entities.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        uris: List[str]
            List of URI of concepts. **Remark:** More than 100 entities are not possible in one request
        force: bool
            Force deletion process

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code
        """
        if len(uris) > 100:
            raise WacomServiceException("Please delete less than 100 entities.")
        url: str = f'{self.service_base_url}{WacomKnowledgeService.CONCEPT_ENDPOINT}'
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        params: Dict[str, Any] = {
            'uris': uris,
            'forceDelete': force
        }
        response: Response = requests.delete(url, headers=headers, params=params, verify=self.verify_calls)
        if not response.ok:
            raise WacomServiceException(f'Deletion of concept failed.'
                                        f'Response code:={response.status_code}, exception:= {response.content}')

    def delete_entity(self, auth_key: str, uri: str, force: bool = False):
        """
        Deletes a entity.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        uri: str
            URI of concept
        force: bool
            Force deletion process

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code
        """
        url: str = f'{self.service_base_url}{WacomKnowledgeService.DELETION_ENDPOINT.format(urllib.parse.quote(uri))}'
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        response: Response = requests.delete(url, headers=headers, params={'forceDelete': force},
                                             verify=self.verify_calls)
        if not response.ok:
            raise WacomServiceException('Deletion of concept failed. Response code:={}, exception:= {}'
                                        .format(response.status_code,
                                                response.content))

    def exists(self, auth_key: str, uri: str) -> bool:
        """
        Check if entity exists in knowledge graph.

        Parameters
        ----------
        auth_key: str -
            User token
        uri: str -
            URI for entity

        Returns
        -------
        flag: bool
            Flag if entity does exist
        """
        try:
            obj: ThingObject = self.entity(auth_key, uri)
            return obj is not None
        except WacomServiceException:
            return False

    def create_entity(self, auth_key: str, concept: ThingObject):
        """
        Creates concept in graph.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        concept: ThingObject
            Concept object

        Returns
        -------
        uri: str
            URI of entity

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code
        """
        url: str = f'{self.service_base_url}{WacomKnowledgeService.CONCEPT_ENDPOINT}'
        # Different localized content
        labels: List[dict] = []
        descriptions: List[dict] = []
        literals: List[dict] = []
        # Header info
        headers: dict = {
            'Content-Type': 'application/json',
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        # Add description in different languages
        for desc in concept.description:
            if len(desc.content) > 0 and not desc.content == ' ':
                descriptions.append({
                    DESCRIPTION_TAG: desc.content,
                    LANGUAGE_CODE_TAG: desc.language_code
                })
        if len(descriptions) == 0:
            #  Adding an empty description
            for label in concept.label:
                if len(label.content) > 0 and not label.content == ' ':
                    descriptions.append({
                        DESCRIPTION_TAG: f'Description of {label.content}',
                        LANGUAGE_CODE_TAG: label.language_code
                    })

        # Labels are tagged as main label
        for label in concept.label:
            if len(label.content) > 0 and label.content != " ":
                labels.append({
                    'value': label.content,
                    LANGUAGE_CODE_TAG: label.language_code,
                    'isMain': True
                })
        # Alias are no main labels
        for label in concept.alias:
            if len(label.content) > 0 and label.content != " ":
                labels.append({
                    'value': label.content,
                    LANGUAGE_CODE_TAG: label.language_code,
                    'isMain': False
                })
        # Labels are tagged as main label
        for literal_property, list_literals in concept.data_properties.items():
            for li in list_literals:
                if li.data_property_type:
                    literals.append({
                        'value': li.value,
                        LANGUAGE_CODE_TAG: li.language_code,
                        DATA_PROPERTY_TAG: li.data_property_type.iri
                    })
        payload: dict = {
            TYPE_TAG: concept.concept_type.iri,
            DESCRIPTIONS_TAG: descriptions,
            IMAGE_TAG: concept.image,
            LABELS_TAG: labels,
            DATA_PROPERTIES_TAG: literals
        }
        if concept.tenant_access_right:
            payload[TENANT_RIGHTS_TAG] = concept.tenant_access_right.to_list()
        response: Response = requests.post(url, json=payload, headers=headers, verify=self.verify_calls)
        if response.ok:
            uri: str = response.json()[URI_TAG]
            return uri

        import json
        raise WacomServiceException(f'Pushing concept failed. '
                                    f'Response code:={response.status_code}, exception:= {response.content}. '
                                    f'Payload: \n{json.dumps(payload, indent=4)}')

    def relations(self, auth_key: str, uri: str) -> Dict[OntologyPropertyReference, ObjectProperty]:
        """
        Retrieve the relations (object properties) of an entity.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        uri: str
            Entity URI of the source

        Returns
        -------
        relations: Dict[OntologyPropertyReference, ObjectProperty]
            All relations a dict

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code
        """
        url: str = f'{self.service_base_url}{WacomKnowledgeService.RELATIONS_ENDPOINT.format(urllib.parse.quote(uri))}'
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        response: Response = requests.get(url, headers=headers, verify=self.verify_calls)
        if response.ok:
            rel: list = response.json().get(RELATIONS_TAG)
            return ObjectProperty.create_from_list(rel)
        raise WacomServiceException(f'Failed to pull relations. '
                                    f'Response code:={response.status_code}, exception:= {response.content}')

    def labels(self, auth_key: str, uri: str) -> List[Label]:
        """
        Extract list labels of entity.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        uri: str
            Entity URI of the source

        Returns
        -------
        labels: List[Label]
            List of labels of an entity.

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code
        """
        url: str = f'{self.service_base_url}{WacomKnowledgeService.LABELS_ENDPOINT.format(urllib.parse.quote(uri))}'
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        response: Response = requests.get(url, headers=headers, verify=self.verify_calls)
        if response.ok:
            response_dict: dict = response.json()
            if LABELS_TAG in response_dict:
                return [Label.create_from_dict(label) for label in response_dict[LABELS_TAG]]
            return []
        raise WacomServiceException('Failed to pull literals. Response code:={}, exception:= {}'
                                    .format(response.status_code, response.content))

    def literals(self, auth_key: str, uri: str) -> List[DataProperty]:
        """
        Collect all literals of entity.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        uri: str
            Entity URI of the source

        Returns
        -------
        labels: List[DataProperty]
            List of data properties of an entity.

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code
        """
        url: str = f'{self.service_base_url}{WacomKnowledgeService.LITERAL_ENDPOINT.format(urllib.parse.quote(uri))}'
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        response: Response = requests.get(url, headers=headers, verify=self.verify_calls)
        if response.ok:
            literals: list = response.json().get(DATA_PROPERTIES_TAG)
            return DataProperty.create_from_list(literals)
        raise WacomServiceException('Failed to pull literals. Response code:={}, exception:= {}'
                                    .format(response.status_code, response.content))

    def create_relation(self, auth_key: str, source: str, relation: OntologyPropertyReference, target: str):
        """
        Creates a relation for a entity to a source entity.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        source: str
            Entity URI of the source
        relation: OntologyPropertyReference
            ObjectProperty property
        target: str
            Entity URI of the target

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code
        """
        url: str = f'{self.service_base_url}{WacomKnowledgeService.RELATION_ENDPOINT.format(source)}'
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        params: dict = {
            RELATION_TAG: relation.iri,
            TARGET: target
        }
        response: Response = requests.post(url, params=params, headers=headers, verify=self.verify_calls)
        if not response.ok:
            import json
            raise WacomServiceException(f'Create relations failed. '
                                        f'Response code:={response.status_code}, exception:= {response.content}. '
                                        f'URL: {url}'
                                        f'Parameters: \n{json.dumps(params, indent=4)}')

    def remove_relation(self, auth_key: str, source: str, relation: OntologyPropertyReference, target: str):
        """
        Removes a relation.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        source: str
            Entity uri of the source
        relation: OntologyPropertyReference
            ObjectProperty property
        target: str
            Entity uri of the target

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code
        """
        url: str = f'{self.service_base_url}' \
                   f'{WacomKnowledgeService.RELATION_ENDPOINT.format(urllib.parse.quote(source))}'
        params: Dict[str, str] = {
            RELATION_TAG: relation.iri,
            TARGET: target
        }
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        # Get response
        response: Response = requests.delete(url, params=params, headers=headers, verify=self.verify_calls)
        if not response.ok:
            raise WacomServiceException(f'Deletion of relation failed. '
                                        f'Response code:={response.status_code}, exception:= {response.content}')

    def activations(self, auth_key: str, uris: List[str], depth: int) \
            -> Tuple[Dict[str, ThingObject], List[Tuple[str, OntologyPropertyReference, str]]]:
        """
        Spreading activation, retrieving the entities related to a  entity.

        Parameters
        ----------
        auth_key: str
            Auth key for user
        uris: List[str]
            List of URIS for entity.
        depth: int
            Depth of activations

        Returns
        -------
        entity_map: Dict[str, ThingObject]
            Map with entity and its URI as key.
        relations: List[Tuple[str, OntologyPropertyReference, str]]
            List of relations with subject predicate, (Property), and subject

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code, and activation failed.
        """
        url: str = f'{self.service_url}/{self.service_endpoint}{WacomKnowledgeService.ACTIVATIONS_ENDPOINT}'

        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        params: dict = {
            'uris': uris,
            'activation': depth
        }

        response: Response = requests.get(url, headers=headers, params=params, verify=self.verify_calls)
        if response.ok:
            entities: Dict[str, Any] = response.json()
            things: Dict[str, ThingObject] = dict([(e[URI_TAG], ThingObject.from_dict(e))
                                                   for e in entities['entities']])
            relations: List[Tuple[str, OntologyPropertyReference, str]] = []
            for r in entities[RELATIONS_TAG]:
                relation: OntologyPropertyReference = OntologyPropertyReference.parse(r[PREDICATE])
                relations.append((r[SUBJECT], relation, r[OBJECT]))
                if r[SUBJECT] in things:
                    things[r[SUBJECT]].add_relation(ObjectProperty(relation, outgoing=[r[OBJECT]]))
            return things, relations

    def listing(self, auth_key: str, filter_type: OntologyClassReference, page_id: Optional[str] = None,
                limit: int = 30, language_code: LanguageCode = 'en_US') -> Tuple[List[ThingObject], int, str]:
        """
        List all entities visible to users.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        filter_type: OntologyClassReference
            Filtering with concept
        page_id: Optional[str]
            Page id. Start from this page id
        limit: int
            Limit of the returned entities
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.

        Returns
        -------
        entities: List[ThingObject]
            List of entities
        total_number: int
            Number of all entities
        next_page_id: str
            Identifier of the next page

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code
        """
        url: str = f'{self.service_url}/{self.service_endpoint}{WacomKnowledgeService.LISTING_ENDPOINT}'
        # Header with auth token
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        # Parameter with filtering and limit
        parameters: dict = {
            TYPE_TAG: filter_type.iri,
            LIMIT_PARAMETER: limit,
            LANGUAGE_PARAMETER: language_code
        }
        # If filtering is configured
        if page_id is not None:
            parameters[NEXT_PAGE_ID_TAG] = page_id
        # Send request
        response: Response = requests.get(url, params=parameters, headers=headers, verify=self.verify_calls)
        # If response is successful
        if response.ok:
            entities_resp: dict = response.json()
            next_page_id: str = entities_resp[NEXT_PAGE_ID_TAG]
            total_number: int = entities_resp[TOTAL_COUNT]
            entities: List[ThingObject] = []
            if LISTING in entities_resp:
                for e in entities_resp[LISTING]:
                    thing: ThingObject = ThingObject.from_dict(e)
                    thing.status_flag = EntityStatus.SYNCED
                    entities.append(thing)
            return entities, total_number, next_page_id
        else:
            raise WacomServiceException('Failed to list the entities (since:= {}, limit:={}). '
                                        'Response code:={}, exception:= {}'
                                        .format(page_id, limit, response.status_code, response.content))

    def ontology_update(self, auth_key: str):
        """
        Update the ontology.

        **Remark:**
        Works for users with role 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code and commit failed.
        """
        url: str = f'{self.service_base_url}{WacomKnowledgeService.ONTOLOGY_UPDATE_ENDPOINT}'
        # Header with auth token
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        response: Response = requests.patch(url, headers=headers, verify=self.verify_calls)
        if not response.ok:
            raise WacomServiceException(f'Ontology update fails. '
                                        f'Response code:={response.status_code}, exception:= {response.content}')

    def search_all(self, auth_key: str, search_term: str, language_code: LanguageCode,
                   types: List[OntologyClassReference], limit: int = 30, next_page_id: str = None)\
            -> Tuple[List[ThingObject], str]:
        """Search term in labels, literals and description.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        search_term: str
            Search term.
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        types: List[OntologyClassReference]
            Limits the types for search.
        limit: int  (default:= 30)
            Size of the page for pagination.
        next_page_id: str (default:=None)
            ID of the next page within pagination.

        Returns
        -------
        results: List[ThingObject]
            List of things matching the search term
        next_page_id: str
            ID of the next page.

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code.
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG:  f'Bearer {auth_key}'
        }
        parameters: Dict[str, Any] = {
            SEARCH_TERM: search_term,
            LANGUAGE_PARAMETER: language_code,
            TYPES_PARAMETER: [ot.iri for ot in types],
            LIMIT: limit,
            NEXT_PAGE_ID_TAG: next_page_id
        }
        url: str = f'{self.service_url}/{self.service_endpoint}{WacomKnowledgeService.SEARCH_TYPES_ENDPOINT}'
        response: Response = requests.get(url, headers=headers, params=parameters, verify=self.verify_calls)
        if response.ok:
            return WacomKnowledgeService.__search_results__(response.json())
        raise WacomServiceException(f'Search on labels {search_term} failed. '
                                    f'{parameters}'
                                    f'Response code:={response.status_code}, exception:= {response.content}')

    def search_labels(self, auth_key: str, search_term: str, language_code: LanguageCode, limit: int = 30,
                      next_page_id: str = None) -> Tuple[List[ThingObject], str]:
        """Search for matches in labels.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        search_term: str
            Search term.
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        limit: int  (default:= 30)
            Size of the page for pagination.
        next_page_id: str (default:=None)
            ID of the next page within pagination.

        Returns
        -------
        results: List[ThingObject]
            List of things matching the search term
        next_page_id: str
            ID of the next page.

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code.
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG:  f'Bearer {auth_key}'
        }
        parameters: Dict[str, Any] = {
            SEARCH_TERM: search_term,
            LANGUAGE_PARAMETER: language_code,
            LIMIT: limit,
            NEXT_PAGE_ID_TAG: next_page_id
        }
        url: str = f'{self.service_url}/{self.service_endpoint}{WacomKnowledgeService.SEARCH_LABELS_ENDPOINT}'
        response: Response = requests.get(url, headers=headers, params=parameters, verify=self.verify_calls)
        if response.ok:
            return WacomKnowledgeService.__search_results__(response.json())
        raise WacomServiceException(f'Search on labels {search_term} failed. '
                                    f'Response code:={response.status_code}, exception:= {response.content}')

    def search_literal(self, auth_key: str, search_term: str,  literal: OntologyPropertyReference,
                       pattern: SearchPattern = SearchPattern.REGEX,
                       language_code: LanguageCode = LanguageCode('en_US'),
                       limit: int = 30, next_page_id: str = None) -> Tuple[List[ThingObject], str]:
        """
        Search for matches in literals.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        search_term: str
            Search term.
        literal: OntologyPropertyReference
            Literal used for the search
        pattern: SearchPattern (default:= SearchPattern.REGEX)
            Search pattern. The chosen search pattern must fit the type of the entity.
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        limit: int (default:= 30)
            Size of the page for pagination.
        next_page_id: str (default:=None)
            ID of the next page within pagination.

        Returns
        -------
        results: List[ThingObject]
           List of things matching the search term
       next_page_id: str
           ID of the next page.

       Raises
       ------
       WacomServiceException
           If the graph service returns an error code.
       """
        url: str = f'{self.service_url}/{self.service_endpoint}{WacomKnowledgeService.SEARCH_LITERALS_ENDPOINT}'
        parameters: Dict[str, Any] = {
            'Value': search_term,
            'Literal': literal.iri,
            LANGUAGE_PARAMETER: language_code,
            LIMIT_PARAMETER: limit,
            'SearchPattern': pattern.value,
            NEXT_PAGE_ID_TAG: next_page_id
        }
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        response: Response = requests.get(url, headers=headers, params=parameters, verify=self.verify_calls)
        if response.ok:
            return WacomKnowledgeService.__search_results__(response.json())
        raise WacomServiceException(f'Search on labels {search_term} failed. '
                                    f'Response code:={response.status_code}, exception:= {response.content}')

    def search_relation(self, auth_key: str, relation: OntologyPropertyReference,
                        language_code: LanguageCode, subject_uri: str = None, object_uri: str = None,
                        limit: int = 30, next_page_id: str = None) -> Tuple[List[ThingObject], str]:
        """
        Search for matches in literals.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        relation: OntologyPropertyReference
            Search term.
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        subject_uri: str (default:=None)
            URI of the subject
        object_uri: str (default:=None)
            URI of the object
        limit: int (default:= 30)
            Size of the page for pagination.
        next_page_id: str (default:=None)
            ID of the next page within pagination.

        Returns
        -------
        results: List[ThingObject]
           List of things matching the search term
        next_page_id: str
           ID of the next page.

       Raises
       ------
       WacomServiceException
           If the graph service returns an error code.
       """
        url: str = f'{self.service_url}/{self.service_endpoint}{WacomKnowledgeService.SEARCH_RELATION_ENDPOINT}'
        parameters: Dict[str, Any] = {
            SUBJECT_URI: subject_uri,
            RELATION_URI: relation.iri,
            OBJECT_URI: object_uri,
            LANGUAGE_PARAMETER: language_code,
            LIMIT: limit,
            NEXT_PAGE_ID_TAG: next_page_id
        }
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }

        response = requests.get(url, headers=headers, params=parameters, verify=self.verify_calls)
        if response.ok:
            return WacomKnowledgeService.__search_results__(response.json())
        raise WacomServiceException(f'Search on: subject:={subject_uri}, relation {relation.iri}, '
                                    f'object:= {object_uri} failed. '
                                    f'Response code:={response.status_code}, exception:= {response.content}')

    def search_description(self, auth_key: str, search_term: str, language_code: LanguageCode, limit: int = 30,
                           next_page_id: str = None) -> Tuple[List[ThingObject], str]:
        """Search for matches in description.

        Parameters
        ----------
        auth_key: str
            Auth key from user
        search_term: str
            Search term.
        language_code: LanguageCode
            ISO-3166 Country Codes and ISO-639 Language Codes in the format '<language_code>_<country>, e.g., en_US.
        limit: int  (default:= 30)
            Size of the page for pagination.
        next_page_id: str (default:=None)
            ID of the next page within pagination.

        Returns
        -------
        results: List[ThingObject]
            List of things matching the search term
        next_page_id: str
            ID of the next page.

        Raises
        ------
        WacomServiceException
            If the graph service returns an error code.
        """
        url: str = f'{self.service_base_url}{WacomKnowledgeService.SEARCH_DESCRIPTION_ENDPOINT}'
        parameters: Dict[str, Any] = {
            SEARCH_TERM: search_term,
            LANGUAGE_PARAMETER: language_code,
            LIMIT: limit,
            NEXT_PAGE_ID_TAG: next_page_id
        }
        headers: dict = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }

        response = requests.get(url, headers=headers, params=parameters, verify=self.verify_calls)
        if response.ok:
            return WacomKnowledgeService.__search_results__(response.json())
        raise WacomServiceException(f'Search on labels {search_term} failed. '
                                    f'Response code:={response.status_code}, exception:= {response.content}')

    @staticmethod
    def __search_results__(response: Dict[str, Any]) -> Tuple[List[ThingObject], str]:
        results: List[ThingObject] = []
        for elem in response['result']:
            results.append(ThingObject.from_dict(elem))
        return results, response[NEXT_PAGE_ID_TAG]



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.
import urllib.parse
from http import HTTPStatus
from typing import Dict, List, Any, Optional, Tuple

import requests
from requests import Response

from knowledge.base.entity import OntologyContext, LocalizedContent, Label, Comment
from knowledge.base.ontology import OntologyClassReference, OntologyPropertyReference, OntologyProperty, OntologyClass,\
    PropertyType, THING_CLASS, DataPropertyType
from knowledge.services.base import WacomServiceAPIClient, WacomServiceException
from knowledge.services.graph import AUTHORIZATION_HEADER_FLAG

# ------------------------------------------------- Constants ----------------------------------------------------------
BASE_URI_TAG: str = "baseUri"
COMMENTS_TAG: str = "comments"
DOMAIN_TAG: str = "domain"
ICON_TAG: str = "icon"
INVERSE_OF_TAG: str = "inverseOf"
KIND_TAG: str = "kind"
LABELS_TAG: str = "labels"
LANGUAGE_CODE: str = 'languageCode'
NAME_TAG: str = "name"
RANGE_TAG: str = "range"
SUB_CLASS_OF_TAG: str = "subClassOf"
SUB_PROPERTY_OF_TAG: str = "subPropertyOf"
TEXT_TAG: str = 'value'


class OntologyService(WacomServiceAPIClient):
    """
    Ontology API Client
    -------------------
    Client to access the ontology service. Offers the following functionality:
    - Listing class names and property names
    - Create new ontology types
    - Update ontology types

    Parameters
    ----------
    service_url: str
        URL of the service
    service_endpoint: str
        Base endpoint
    """
    CONTEXT_ENDPOINT: str = 'context'
    CONCEPTS_ENDPOINT: str = 'context/{}/concepts'
    CONCEPT_ENDPOINT: str = 'context/{}/concepts/{}'
    PROPERTIES_ENDPOINT: str = "context/{}/properties"
    COMMIT_ENDPOINT: str = "context/{}/commit"
    RDF_ENDPOINT: str = "context/{}/versions/rdf"
    PROPERTY_ENDPOINT: str = "context/{}/properties/{}"

    def __init__(self, service_url: str, service_endpoint: str = 'ontology'):
        super().__init__(application_name="Ontology Service", service_url=service_url,
                         service_endpoint=service_endpoint)

    def contexts(self, auth_key: str) -> List[OntologyContext]:
        """List all concepts.

        **Remark:**
        Works for users with role 'User' and 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.

        Returns
        -------
        contexts: List[OntologyContext]
            List of ontology contexts
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        response: Response = requests.get(f'{self.service_base_url}{OntologyService.CONTEXT_ENDPOINT}',
                                          headers=headers, verify=self.verify_calls)
        if response.ok:
            response_list: List[OntologyContext] = []
            for c in response.json():
                response_list.append(OntologyContext.from_dict(c))
            return response_list
        raise WacomServiceException(f'Listing of context failed. '
                                    f'Response code:={response.status_code}, exception:= {response.text}')

    def concepts(self, auth_key: str, context: str) -> List[Tuple[OntologyClassReference, OntologyClassReference]]:
        """Retrieve all concept classes.

        **Remark:**
        Works for users with role 'User' and 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.
        context: str
            Context of the ontology

        Returns
        -------
        concepts: List[Tuple[OntologyClassReference, OntologyClassReference]]
            List of ontology classes. Tuple<Classname, Superclass>
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        response: Response = requests.get(f'{self.service_base_url}{OntologyService.CONCEPTS_ENDPOINT.format(context)}',
                                          headers=headers, verify=self.verify_calls)
        if response.ok:
            response_list: List[Tuple[OntologyClassReference, OntologyClassReference]] = []
            result = response.json()
            for struct in result:
                response_list.append((OntologyClassReference.parse(struct[NAME_TAG]),
                                      None if struct[SUB_CLASS_OF_TAG] is None else
                                      OntologyClassReference.parse(struct[SUB_CLASS_OF_TAG])))
            return response_list
        raise WacomServiceException(f'Response code:={response.status_code}, exception:= {response.text}')

    def properties(self, auth_key: str, context_name: str) -> List[Tuple[OntologyPropertyReference,
                                                                         OntologyPropertyReference]]:
        """List all properties.

        **Remark:**
        Works for users with role 'User' and 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.
        context_name: str
            Name of the context

        Returns
        -------
        contexts: List[Tuple[OntologyPropertyReference, OntologyPropertyReference]]
            List of ontology contexts
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        context_url: str = urllib.parse.quote_plus(context_name)

        response: Response = requests.get(f'{self.service_base_url}'
                                          f'{OntologyService.PROPERTIES_ENDPOINT.format(context_url)}',
                                          headers=headers, verify=self.verify_calls)
        # Return empty list if the NOT_FOUND is reported
        if response.status_code == HTTPStatus.NOT_FOUND:
            return []
        if response.ok:
            response_list: List[Tuple[OntologyPropertyReference, OntologyPropertyReference]] = []
            for c in response.json():
                response_list.append((OntologyPropertyReference.parse(c[NAME_TAG]),
                                      None if c[SUB_PROPERTY_OF_TAG] is None else
                                      OntologyPropertyReference.parse(c[SUB_PROPERTY_OF_TAG])))
            return response_list
        raise WacomServiceException(f'Response code:={response.status_code}, exception:= {response.text}')

    def concept(self, auth_key: str, context_name: str, concept_name: str) -> OntologyClass:
        """Retrieve a concept instance.

        **Remark:**
        Works for users with role 'User' and 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.
        context_name: str
            Name of the context
        concept_name: str
            IRI of the concept

        Returns
        -------
        instance: OntologyClass
            Instance of the concept
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        context_url: str = urllib.parse.quote_plus(context_name)
        concept_url: str = urllib.parse.quote_plus(concept_name)
        response: Response = requests.get(f'{self.service_base_url}'
                                          f'{OntologyService.CONCEPT_ENDPOINT.format(context_url, concept_url)}',
                                          headers=headers, verify=self.verify_calls)
        if response.ok:
            result: Dict[str, Any] = response.json()
            return OntologyClass.from_dict(result)
        raise WacomServiceException(f'Response code:={response.status_code}, exception:= {response.text}')

    def property(self, auth_key: str, context_name: str, property_name: str) -> OntologyProperty:
        """Retrieve a property instance.

        **Remark:**
        Works for users with role 'User' and 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.
        context_name: str
            Name of the context
        property_name: str
            IRI of the property

        Returns
        -------
        instance: OntologyProperty
            Instance of the property
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        context_url: str = urllib.parse.quote_plus(context_name)
        concept_url: str = urllib.parse.quote_plus(property_name)
        param: str = OntologyService.PROPERTY_ENDPOINT.format(context_url, concept_url)
        response: Response = requests.get(f'{self.service_base_url}{param}', headers=headers, verify=self.verify_calls)
        if response.ok:
            return OntologyProperty.from_dict(response.json())
        raise WacomServiceException(f'Response code:={response.status_code}, exception:= {response.text}')

    def create_concept(self, auth_key: str, context: str, reference: OntologyClassReference,
                       subclass_of: OntologyClassReference = THING_CLASS,
                       icon: Optional[str] = None, labels: Optional[List[Label]] = None,
                       comments: Optional[List[Comment]] = None) -> Dict[str, str]:
        """Create concept class.

        **Remark:**
        Only works for users with role 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.
        context: str
            Context of ontology
        reference: OntologyClassReference
            Name of the concept
        subclass_of: OntologyClassReference (default:=wacom:core#Thing)
            Super class of the concept
        icon: Optional[str] (default:= None)
            Icon representing the concept
        labels: Optional[List[Label]] (default:= None)
            Labels for the class
        comments: Optional[List[Comment]] (default:= None)
            Comments for the class
        Returns
        -------
        result: Dict[str, str]
            Result from the service

        Raises
        ------
        WacomServiceException
            If the ontology service returns an error code, exception is thrown.
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        payload: Dict[str, Any] = {
            SUB_CLASS_OF_TAG: subclass_of.iri,
            NAME_TAG: reference.iri,
            LABELS_TAG: [],
            COMMENTS_TAG: [],
            ICON_TAG: icon
        }
        context_url: str = urllib.parse.quote_plus(context)
        for label in labels if labels is not None else []:
            payload[LABELS_TAG].append({TEXT_TAG: label.content, LANGUAGE_CODE: label.language_code})
        for comment in comments if comments is not None else []:
            payload[COMMENTS_TAG].append({TEXT_TAG: comment.content, LANGUAGE_CODE: comment.language_code})
        url: str = f'{self.service_base_url}{OntologyService.CONCEPTS_ENDPOINT.format(context_url)}'
        response: Response = requests.post(url, headers=headers, json=payload, verify=self.verify_calls)
        if response.ok:
            result_dict: Dict[str, str] = response.json()
            return result_dict
        raise WacomServiceException(f'Creation of concept failed. '
                                    f'Response code:={response.status_code}, exception:= {response.text}')

    def update_concept(self, auth_key: str, context: str, name: str, subclass_of: Optional[str],
                       icon: Optional[str] = None, labels: Optional[List[Label]] = None,
                       comments: Optional[List[Comment]] = None) -> Dict[str, str]:
        """Update concept class.

        **Remark:**
        Only works for users with role 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.
        context: str
            Context of ontology
        name: str
            Name of the concept
        subclass_of: Optional[str]
            Super class of the concept
        icon: Optional[str] (default:= None)
            Icon representing the concept
        labels: Optional[List[Label]] (default:= None)
            Labels for the class
        comments: Optional[List[Comment]] (default:= None)
            Comments for the class

        Returns
        -------
        response: Dict[str, str]
            Response from service

        Raises
        ------
        WacomServiceException
            If the ontology service returns an error code, exception is thrown.
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        payload: Dict[str, Any] = {
            SUB_CLASS_OF_TAG: subclass_of,
            NAME_TAG: name,
            LABELS_TAG: [],
            COMMENTS_TAG: [],
            ICON_TAG: icon
        }
        context_url: str = urllib.parse.quote_plus(context)
        for label in labels if labels is not None else []:
            payload[LABELS_TAG].append({TEXT_TAG: label.content, LANGUAGE_CODE: label.language_code})
        for comment in comments if comments is not None else []:
            payload[COMMENTS_TAG].append({TEXT_TAG: comment.content, LANGUAGE_CODE: comment.language_code})
        url: str = f'{self.service_url}{self.service_endpoint}' \
                   f'{OntologyService.CONCEPTS_ENDPOINT.format(context_url)}'
        response: Response = requests.put(url, headers=headers, json=payload, verify=self.verify_calls)
        if response.ok:
            return response.json()
        raise WacomServiceException(f'Update of concept failed. '
                                    f'Response code:={response.status_code}, exception:= {response.text}')

    def create_object_property(self, auth_key: str, context: str,
                               reference: OntologyPropertyReference,
                               domain_cls: OntologyClassReference, range_cls: OntologyClassReference,
                               inverse_of: Optional[OntologyPropertyReference] = None,
                               subproperty_of: Optional[OntologyPropertyReference] = None,
                               icon: Optional[str] = None,
                               labels: Optional[List[LocalizedContent]] = None,
                               comments: Optional[List[LocalizedContent]] = None) -> Dict[str, str]:
        """Create property.

        **Remark:**
        Only works for users with role 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.
        context: str
            Context of ontology
        reference: OntologyPropertyReference
            Name of the concept
        domain_cls: OntologyClassReference
            IRI of the domain
        range_cls: OntologyClassReference
            IRI of the range
        inverse_of: Optional[OntologyPropertyReference] (default:= None)
            Inverse property
        subproperty_of: Optional[OntologyPropertyReference] = None,
            Super property of the concept
        icon: Optional[str] (default:= None)
            Icon representing the concept
        labels: Optional[List[Label]] (default:= None)
            Labels for the class
        comments: Optional[List[Comment]] (default:= None)
            Comments for the class

        Returns
        -------
        result: Dict[str, str]
            Result from the service

        Raises
        ------
        WacomServiceException
            If the ontology service returns an error code, exception is thrown.
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        payload: Dict[str, Any] = {
            KIND_TAG: PropertyType.OBJECT_PROPERTY.value,
            DOMAIN_TAG: domain_cls.iri,
            RANGE_TAG: range_cls.iri,
            SUB_PROPERTY_OF_TAG: subproperty_of.iri if subproperty_of is not None else None,
            INVERSE_OF_TAG: inverse_of.iri if inverse_of is not None else None,
            NAME_TAG: reference.iri,
            LABELS_TAG: [],
            COMMENTS_TAG: [],
            ICON_TAG: icon
        }
        context_url: str = urllib.parse.quote_plus(context)
        for label in labels if labels is not None else []:
            payload[LABELS_TAG].append({TEXT_TAG: label.content, LANGUAGE_CODE: label.language_code})
        for comment in comments if comments is not None else []:
            payload[COMMENTS_TAG].append({TEXT_TAG: comment.content, LANGUAGE_CODE: comment.language_code})
        url: str = f'{self.service_base_url}{OntologyService.PROPERTIES_ENDPOINT.format(context_url)}'
        response: Response = requests.post(url, headers=headers, json=payload, verify=self.verify_calls)
        if response.ok:
            return response.json()
        raise WacomServiceException(f'Creation of object property failed. '
                                    f'Response code:={response.status_code}, exception:= {response.text}')

    def create_data_property(self, auth_key: str, context: str,
                             reference: OntologyPropertyReference,
                             domain_cls: OntologyClassReference, range_cls: DataPropertyType,
                             subproperty_of: Optional[OntologyPropertyReference] = None,
                             icon: Optional[str] = None,
                             labels: Optional[List[LocalizedContent]] = None,
                             comments: Optional[List[LocalizedContent]] = None) -> Dict[str, str]:
        """Create data property.

        **Remark:**
        Only works for users with role 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.
        context: str
            Context of ontology
        reference: OntologyPropertyReference
            Name of the concept
        domain_cls: OntologyClassReference
            IRI of the domain
        range_cls: DataPropertyType
            Data property type
        subproperty_of: Optional[OntologyPropertyReference] = None,
            Super property of the concept
        icon: Optional[str] (default:= None)
            Icon representing the concept
        labels: Optional[List[Label]] (default:= None)
            Labels for the class
        comments: Optional[List[Comment]] (default:= None)
            Comments for the class

        Returns
        -------
        result: Dict[str, str]
            Result from the service

        Raises
        ------
        WacomServiceException
            If the ontology service returns an error code, exception is thrown.
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        payload: Dict[str, Any] = {
            KIND_TAG: PropertyType.DATA_PROPERTY.value,
            DOMAIN_TAG: domain_cls.iri,
            RANGE_TAG: range_cls.value,
            SUB_PROPERTY_OF_TAG: subproperty_of.iri if subproperty_of is not None else None,
            NAME_TAG: reference.iri,
            LABELS_TAG: [],
            COMMENTS_TAG: [],
            ICON_TAG: icon
        }
        context_url: str = urllib.parse.quote_plus(context)
        for label in labels if labels is not None else []:
            payload[LABELS_TAG].append({TEXT_TAG: label.content, LANGUAGE_CODE: label.language_code})
        for comment in comments if comments is not None else []:
            payload[COMMENTS_TAG].append({TEXT_TAG: comment.content, LANGUAGE_CODE: comment.language_code})
        url: str = f'{self.service_base_url}{OntologyService.PROPERTIES_ENDPOINT.format(context_url)}'
        response: Response = requests.post(url, headers=headers, json=payload, verify=self.verify_calls)
        if response.ok:
            return response.json()
        raise WacomServiceException(f'Creation of data property failed. '
                                    f'Response code:={response.status_code}, exception:= {response.text}')

    def create_context(self, auth_key: str, name: str, base_uri: Optional[str] = None, icon: Optional[str] = None,
                       labels: List[Label] = None, comments: List[Comment] = None) -> Dict[str, str]:
        """Create context.

        **Remark:**
        Only works for users with role 'TenantAdmin'.

        Parameters
        ----------
        auth_key: str
            Auth key from user.
        base_uri: str
            Base URI
        name: str
            Name of the context
        icon: Optional[str] (default:= None)
            Icon representing the concept
        labels: Optional[List[Label]] (default:= None)
            Labels for the context
        comments: Optional[List[Comment]] (default:= None)
            Comments for the context

        Returns
        -------
        result: Dict[str, str]
            Result from the service

        Raises
        ------
        WacomServiceException
            If the ontology service returns an error code, exception is thrown.
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        payload: Dict[str, Any] = {
            BASE_URI_TAG: base_uri if base_uri is not None else f'wacom:{name}',
            NAME_TAG: name,
            LABELS_TAG: [],
            COMMENTS_TAG: [],
            ICON_TAG: icon
        }
        for label in labels if labels is not None else []:
            payload[LABELS_TAG].append({TEXT_TAG: label.content, LANGUAGE_CODE: label.language_code})
        for comment in comments if comments is not None else []:
            payload[COMMENTS_TAG].append({TEXT_TAG: comment.content, LANGUAGE_CODE: comment.language_code})
        url: str = f'{self.service_base_url}{OntologyService.CONTEXT_ENDPOINT}'
        response: Response = requests.post(url, headers=headers, json=payload, verify=self.verify_calls)
        if response.ok:
            return response.json()
        raise WacomServiceException(f'Creation of concept failed. '
                                    f'Response code:={response.status_code}, exception:= {response.text}')

    def commit(self, auth_key: str, context_name: str):
        """
        Commit the ontology.

        Parameters
        ----------
        auth_key: str
            User token (must have TenantAdmin) role
        context_name: str
            Name of the context.
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        context_url: str = urllib.parse.quote_plus(context_name)
        url: str = f'{self.service_base_url}{OntologyService.COMMIT_ENDPOINT.format(context_url)}'
        response: Response = requests.put(url, headers=headers, verify=self.verify_calls)
        if not response.ok:
            raise WacomServiceException(f'Commit of ontology failed. '
                                        f'Response code:={response.status_code}, exception:= {response.text}')

    def rdf_export(self, auth_key: str, context_name: str) -> str:
        """
        Export RDF.

        Parameters
        ----------
        auth_key: str
            User token (must have TenantAdmin) role
        context_name: str
            Name of the context.

        Returns
        -------
        rdf: str
            Ontology as RDFS / OWL  ontology
        """
        headers: Dict[str, str] = {
            AUTHORIZATION_HEADER_FLAG: f'Bearer {auth_key}'
        }
        context_url: str = urllib.parse.quote_plus(context_name)
        url: str = f'{self.service_base_url}{OntologyService.RDF_ENDPOINT.format(context_url)}'
        response: Response = requests.get(url, headers=headers, verify=self.verify_calls)
        if response.ok:
            return response.text
        raise WacomServiceException(f'RDF export failed. '
                                    f'Response code:={response.status_code}, exception:= {response.text}')



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.

__all__ = ['access', 'entity', 'ontology']

from knowledge.base import access
from knowledge.base import entity
from knowledge.base import ontology



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.

import logging
from typing import Optional

__author__ = "Markus Weber"
__copyright__ = "Copyright 2021 Wacom. All rights reserved."
__version__ = "1.0.0"
__credits__ = ["Markus Weber"]
__license__ = "Wacom"
__maintainer__ = ["Markus Weber"]
__email__ = "markus.weber@wacom.com"
__status__ = "beta"

# Create the Logger
logger: Optional[logging.Logger] = None

if logger is None:
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    ch: logging.StreamHandler = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    # create formatter and add it to the handlers
    formatter: logging.Formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(ch)

__all__ = ['__copyright__', '__credits__', '__license__', '__maintainer__', '__email__', '__status__', '__version__',
           'logger', 'base', 'nel', 'public', 'services', 'utils']

from knowledge import base
from knowledge import nel
from knowledge import public
from knowledge import services
from knowledge import utils
